import inspect
import logging
import tempfile
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

import libcst as cst
import numpy as np
from IPython.display import display
from PIL import Image

import vision_agent.tools as T
from vision_agent.agent.agent_utils import DefaultImports, extract_json, extract_tag
from vision_agent.agent.vision_agent_planner_prompts_v2 import (
    CATEGORIZE_TOOL_REQUEST,
    FINALIZE_PLAN,
    PICK_TOOL,
    TEST_TOOLS,
    TEST_TOOLS_EXAMPLE1,
    TEST_TOOLS_EXAMPLE2,
)
from vision_agent.configs import Config
from vision_agent.lmm import LMM, AnthropicLMM
from vision_agent.utils.execute import (
    CodeInterpreter,
    CodeInterpreterFactory,
    Execution,
    MimeType,
)
from vision_agent.utils.image_utils import convert_to_b64
from vision_agent.utils.sim import get_tool_recommender

TOOL_FUNCTIONS = {tool.__name__: tool for tool in T.TOOLS}
LOAD_TOOLS_DOCSTRING = T.get_tool_documentation(
    [T.load_image, T.extract_frames_and_timestamps]
)

CONFIG = Config()
_LOGGER = logging.getLogger(__name__)
EXAMPLES = f"\n{TEST_TOOLS_EXAMPLE1}\n{TEST_TOOLS_EXAMPLE2}\n"


def format_tool_output(tool_thoughts: str, tool_docstring: str) -> str:
    return_str = "[get_tool_for_task output]\n"
    if tool_thoughts.strip() != "":
        return_str += f"{tool_thoughts}\n\n"
    return_str += (
        f"Tool Documentation:\n{tool_docstring}\n[end of get_tool_for_task output]\n"
    )
    return return_str


def run_multi_judge(
    tool_chooser: LMM,
    tool_docs_str: str,
    task: str,
    code: str,
    tool_output_str: str,
    image_paths: List[str],
) -> Tuple[Optional[Callable], str, str]:
    error_message = ""
    prompt = PICK_TOOL.format(
        tool_docs=tool_docs_str,
        user_request=task,
        context=f"<code>\n{code}\n</code>\n<tool_output>\n{tool_output_str}\n</tool_output>",
        previous_attempts=error_message,
    )

    def run_judge() -> Tuple[Optional[Callable], str, str]:
        response = tool_chooser.generate(prompt, media=image_paths, temperature=1.0)
        tool_choice_context = extract_tag(response, "json")  # type: ignore
        tool_choice_context_dict = extract_json(tool_choice_context)  # type: ignore
        tool, tool_thoughts, tool_docstring, _ = extract_tool_info(
            tool_choice_context_dict
        )
        return tool, tool_thoughts, tool_docstring

    responses = []
    with ThreadPoolExecutor() as executor:
        futures = [executor.submit(run_judge) for _ in range(3)]
        for future in as_completed(futures):
            responses.append(future.result())

    responses = [r for r in responses if r[0] is not None]
    counts: Dict[str, int] = {}
    for tool, tool_thoughts, tool_docstring in responses:
        if tool is not None:
            counts[tool.__name__] = counts.get(tool.__name__, 0) + 1
            if counts[tool.__name__] >= 2:
                return tool, tool_thoughts, tool_docstring

    if len(responses) == 0:
        return (
            None,
            "No tool could be found, please try again with a different prompt or image",
            "",
        )
    return responses[0]


def extract_tool_info(
    tool_choice_context: Dict[str, Any],
) -> Tuple[Optional[Callable], str, str, str]:
    tool_thoughts = tool_choice_context.get("thoughts", "")
    tool_docstring = ""
    tool = tool_choice_context.get("best_tool", None)
    if tool in TOOL_FUNCTIONS:
        tool = TOOL_FUNCTIONS[tool]
        tool_docstring = T.TOOLS_INFO[tool.__name__]

    return tool, tool_thoughts, tool_docstring, ""


def replace_box_threshold(code: str, functions: List[str], box_threshold: float) -> str:
    class ReplaceBoxThresholdTransformer(cst.CSTTransformer):
        def leave_Call(
            self, original_node: cst.Call, updated_node: cst.Call
        ) -> cst.Call:
            if (
                isinstance(updated_node.func, cst.Name)
                and updated_node.func.value in functions
            ) or (
                isinstance(updated_node.func, cst.Attribute)
                and updated_node.func.attr.value in functions
            ):
                new_args = []
                found = False
                for arg in updated_node.args:
                    if arg.keyword and arg.keyword.value == "box_threshold":
                        new_arg = arg.with_changes(value=cst.Float(str(box_threshold)))
                        new_args.append(new_arg)
                        found = True
                    else:
                        new_args.append(arg)

                if not found:
                    new_args.append(
                        cst.Arg(
                            keyword=cst.Name("box_threshold"),
                            value=cst.Float(str(box_threshold)),
                            equal=cst.AssignEqual(
                                whitespace_before=cst.SimpleWhitespace(""),
                                whitespace_after=cst.SimpleWhitespace(""),
                            ),
                        )
                    )
                return updated_node.with_changes(args=new_args)
            return updated_node

    tree = cst.parse_module(code)
    transformer = ReplaceBoxThresholdTransformer()
    new_tree = tree.visit(transformer)
    return new_tree.code


def run_tool_testing(
    task: str,
    image_paths: List[str],
    lmm: LMM,
    exclude_tools: Optional[List[str]],
    code_interpreter: CodeInterpreter,
    process_code: Callable[[str], str] = lambda x: x,
) -> tuple[str, str, Execution]:
    """Helper function to generate and run tool testing code."""
    query = lmm.generate(CATEGORIZE_TOOL_REQUEST.format(task=task))
    category = extract_tag(query, "category")  # type: ignore
    if category is None:
        query = task
    else:
        query = f"{category.strip()}. {task}"

    tool_docs = get_tool_recommender().top_k(query, k=5, thresh=0.3)
    if exclude_tools is not None and len(exclude_tools) > 0:
        cleaned_tool_docs = []
        for tool_doc in tool_docs:
            if not tool_doc["name"] in exclude_tools:
                cleaned_tool_docs.append(tool_doc)
        tool_docs = cleaned_tool_docs
    tool_docs_str = "\n".join([e["doc"] for e in tool_docs])
    tool_docs_str += "\n" + LOAD_TOOLS_DOCSTRING

    prompt = TEST_TOOLS.format(
        tool_docs=tool_docs_str,
        previous_attempts="",
        user_request=task,
        examples=EXAMPLES,
        media=str(image_paths),
    )

    response = lmm.generate(prompt, media=image_paths)
    code = extract_tag(response, "code")  # type: ignore
    if code is None:
        raise ValueError(f"Could not extract code from response: {response}")

    # If there's a syntax error with the code, process_code can crash. Executing the
    # code and then sending the error to the LLM should correct it.
    try:
        code = process_code(code)
    except Exception as e:
        _LOGGER.error(f"Error processing code: {e}")

    tool_output = code_interpreter.exec_isolation(DefaultImports.prepend_imports(code))
    tool_output_str = tool_output.text(include_results=False).strip()

    count = 1
    while (
        not tool_output.success
        or (len(tool_output.logs.stdout) == 0 and len(tool_output.logs.stderr) == 0)
    ) and count <= 3:
        if tool_output_str.strip() == "":
            tool_output_str = "EMPTY"
        prompt = TEST_TOOLS.format(
            tool_docs=tool_docs_str,
            previous_attempts=f"<code>\n{code}\n</code>\nTOOL OUTPUT\n{tool_output_str}",
            user_request=task,
            examples=EXAMPLES,
            media=str(image_paths),
        )
        response = cast(str, lmm.generate(prompt, media=image_paths))
        code = extract_tag(response, "code")
        if code is None:
            code = response

        try:
            code = process_code(code)
        except Exception as e:
            _LOGGER.error(f"Error processing code: {e}")
        tool_output = code_interpreter.exec_isolation(
            DefaultImports.prepend_imports(code)
        )
        tool_output_str = tool_output.text(include_results=False).strip()
        count += 1

    return code, tool_docs_str, tool_output


def get_tool_for_task(
    task: str,
    images: Union[Dict[str, List[np.ndarray]], List[np.ndarray]],
    exclude_tools: Optional[List[str]] = None,
) -> None:
    """Given a task and one or more images this function will find a tool to accomplish
    the jobs. It prints the tool documentation and thoughts on why it chose the tool.

    It can produce tools for the following types of tasks:
        - Object detection and counting
        - Classification
        - Segmentation
        - OCR
        - VQA
        - Depth and pose estimation
        - Video object tracking
        - Video temporal localization (action recognition)
        - Image inpainting

    Only ask for one type of task at a time, for example a task needing to identify
    text is one OCR task while needing to identify non-text objects is an OD task. Wait
    until the documentation is printed to use the function so you know what the input
    and output signatures are.

    Parameters:
        task: str: The task to accomplish.
        images: Union[Dict[str, List[np.ndarray]], List[np.ndarray]]: The images to use
            for the task. If a key is provided, it is used as the file name.
        exclude_tools: Optional[List[str]]: A list of tool names to exclude from the
            recommendations. This is helpful if you are calling get_tool_for_task twice
            and do not want the same tool recommended.

    Returns:
        The tool to use for the task is printed to stdout

    Examples
    --------
        >>> get_tool_for_task(
        >>>     "Give me an OCR model that can find 'hot chocolate' in the image",
        >>>     {"image": [image]})
        >>> get_tool_for_taks(
        >>>     "I need a tool that can paint a background for this image and maks",
        >>>     {"image": [image], "mask": [mask]})
    """
    tool_tester = CONFIG.create_tool_tester()
    tool_chooser = CONFIG.create_tool_chooser()

    if isinstance(images, list):
        images = {"image": images}

    with (
        tempfile.TemporaryDirectory() as tmpdirname,
        CodeInterpreterFactory.new_instance() as code_interpreter,
    ):
        image_paths = []
        for k in images.keys():
            for i, image in enumerate(images[k]):
                image_path = f"{tmpdirname}/{k}_{i}.png"
                Image.fromarray(image).save(image_path)
                image_paths.append(image_path)

        code, tool_docs_str, tool_output = run_tool_testing(
            task, image_paths, tool_tester, exclude_tools, code_interpreter
        )
        tool_output_str = tool_output.text(include_results=False).strip()

        _, tool_thoughts, tool_docstring = run_multi_judge(
            tool_chooser, tool_docs_str, task, code, tool_output_str, image_paths
        )

    print(format_tool_output(tool_thoughts, tool_docstring))


def get_tool_documentation(tool_name: str) -> str:
    # use same format as get_tool_for_task
    tool_doc = T.TOOLS_DF[T.TOOLS_DF["name"] == tool_name]["doc"].values[0]
    return format_tool_output("", tool_doc)


def get_tool_for_task_human_reviewer(
    task: str,
    images: Union[Dict[str, List[np.ndarray]], List[np.ndarray]],
    exclude_tools: Optional[List[str]] = None,
) -> None:
    # NOTE: this will have the same documentation as get_tool_for_task
    tool_tester = CONFIG.create_tool_tester()

    if isinstance(images, list):
        images = {"image": images}

    with (
        tempfile.TemporaryDirectory() as tmpdirname,
        CodeInterpreterFactory.new_instance() as code_interpreter,
    ):
        image_paths = []
        for k in images.keys():
            for i, image in enumerate(images[k]):
                image_path = f"{tmpdirname}/{k}_{i}.png"
                Image.fromarray(image).save(image_path)
                image_paths.append(image_path)

        tools = [
            t.__name__
            for t in T.TOOLS
            if inspect.signature(t).parameters.get("box_threshold")  # type: ignore
        ]

        _, _, tool_output = run_tool_testing(
            task,
            image_paths,
            tool_tester,
            exclude_tools,
            code_interpreter,
            process_code=lambda x: replace_box_threshold(x, tools, 0.05),
        )

        # need to re-display results for the outer notebook to see them
        for result in tool_output.results:
            if "json" in result.formats():
                display({MimeType.APPLICATION_JSON: result.json}, raw=True)


def check_function_call(code: str, function_name: str) -> bool:
    class FunctionCallVisitor(cst.CSTVisitor):
        def __init__(self) -> None:
            self.function_name = function_name
            self.function_called = False

        def visit_Call(self, node: cst.Call) -> None:
            if (
                isinstance(node.func, cst.Name)
                and node.func.value == self.function_name
            ):
                self.function_called = True

    tree = cst.parse_module(code)
    visitor = FunctionCallVisitor()
    tree.visit(visitor)
    return visitor.function_called


def finalize_plan(user_request: str, chain_of_thoughts: str) -> str:
    """Finalizes the plan by taking the user request and the chain of thoughts that
    represent the plan and returns the finalized plan.
    """
    lmm = AnthropicLMM()
    prompt = FINALIZE_PLAN.format(
        user_request=user_request, chain_of_thoughts=chain_of_thoughts
    )
    finalized_plan = cast(str, lmm.generate(prompt))
    return finalized_plan


def claude35_vqa(prompt: str, medias: List[np.ndarray]) -> None:
    """Asks the Claude-3.5 model a question about the given media and returns an answer.

    Parameters:
        prompt: str: The question to ask the model.
        medias: List[np.ndarray]: The images to ask the question about, it could also
            be frames from a video. You can send up to 5 frames from a video.
    """
    vqa = CONFIG.create_vqa()
    if isinstance(medias, np.ndarray):
        medias = [medias]
    if isinstance(medias, list) and len(medias) > 5:
        medias = medias[:5]
    all_media_b64 = [
        "data:image/png;base64," + convert_to_b64(media) for media in medias
    ]

    response = cast(str, vqa.generate(prompt, media=all_media_b64))
    print(f"[claude35_vqa output]\n{response}\n[end of claude35_vqa output]")


def suggestion(prompt: str, medias: List[np.ndarray]) -> None:
    """Given your problem statement and the images, this will provide you with a
    suggested plan on how to proceed. Always call suggestion when starting to solve
    a problem.

    Parameters:
        prompt: str: The problem statement.
        medias: List[np.ndarray]: The images to use for the problem
    """
    try:
        from .suggestion import suggestion_impl  # type: ignore

        suggestion = suggestion_impl(prompt, medias)
        print(suggestion)
    except ImportError:
        print("")


PLANNER_TOOLS = [
    claude35_vqa,
    suggestion,
    get_tool_for_task,
]
PLANNER_DOCSTRING = T.get_tool_documentation(PLANNER_TOOLS)  # type: ignore



import os
import platform
import shutil
from functools import lru_cache
from importlib import resources
from pathlib import Path
from typing import Callable, Dict, List, Optional, Sequence, Union

import numpy as np
import pandas as pd
import requests
from openai import AzureOpenAI, OpenAI
from scipy.spatial.distance import cosine  # type: ignore

from vision_agent.tools.tool_utils import (
    _LND_API_KEY,
    _create_requests_session,
    _LND_API_URL_v2,
)
from vision_agent.tools.tools import TOOLS_DF


@lru_cache(maxsize=1)
def get_tool_recommender() -> "Sim":
    return load_cached_sim(TOOLS_DF)


@lru_cache(maxsize=512)
def get_embedding(
    emb_call: Callable[[List[str]], List[float]], text: str
) -> List[float]:
    text = text.replace("\n", " ")
    return emb_call([text])


def load_cached_sim(
    tools_df: pd.DataFrame, sim_key: str = "desc", cached_dir: str = ".sim_tools"
) -> "Sim":
    cached_dir_full_path = str(resources.files("vision_agent") / cached_dir)
    if os.path.exists(cached_dir_full_path):
        if tools_df is not None:
            if StellaSim.check_load(cached_dir_full_path, tools_df):
                # don't pass sim_key to loaded Sim object or else it will re-calculate embeddings
                return StellaSim.load(cached_dir_full_path)
    if os.path.exists(cached_dir_full_path):
        shutil.rmtree(cached_dir_full_path)

    sim = StellaSim(tools_df, sim_key=sim_key)
    sim.save(cached_dir_full_path)
    return sim


def stella_embeddings(prompts: List[str]) -> List[np.ndarray]:
    # implement stella here so we don't display embeddings which can be costly
    payload = {
        "input": prompts,
        "model": "stella1.5b",
    }
    url = f"{_LND_API_URL_v2}/embeddings"
    headers = {"apikey": _LND_API_KEY}
    session = _create_requests_session(
        url=url,
        num_retry=3,
        headers=headers,
    )
    response = session.post(url, data=payload).json()
    data = response["data"]
    return [np.array(d["embedding"]) for d in data]


class Sim:
    def __init__(
        self,
        df: pd.DataFrame,
        sim_key: Optional[str] = None,
        api_key: Optional[str] = None,
        model: str = "text-embedding-3-small",
    ) -> None:
        """Creates a similarity object that can be used to find similar items in a
        dataframe.

        Parameters:
            df: pd.DataFrame: The dataframe to use for similarity.
            sim_key: Optional[str]: The column name that you want to use to construct
                the embeddings.
            api_key: Optional[str]: The OpenAI API key to use for embeddings.
            model: str: The model to use for embeddingshttps://github.com/landing-ai/vision-agent/pull/280.
        """
        self.df = df
        self.client = OpenAI(api_key=api_key)
        self.emb_call = (
            lambda x: self.client.embeddings.create(input=x, model=model)
            .data[0]
            .embedding
        )
        self.model = model
        if "embs" not in df.columns and sim_key is None:
            raise ValueError("key is required if no column 'embs' is present.")

        if sim_key is not None:
            self.df = self.df.assign(
                embs=self.df[sim_key].apply(
                    lambda x: get_embedding(
                        self.emb_call,
                        x,
                    )
                )
            )

    def save(self, save_dir: Union[str, Path]) -> None:
        save_dir = Path(save_dir)
        save_dir.mkdir(parents=True, exist_ok=True)

        df = self.df.copy()
        embs = np.array(df.embs.tolist())
        np.save(save_dir / "embs.npy", embs)
        df = df.drop("embs", axis=1)
        df.to_csv(save_dir / "df.csv", index=False)

    @staticmethod
    def load(
        load_dir: Union[str, Path],
        api_key: Optional[str] = None,
        model: str = "text-embedding-3-small",
    ) -> "Sim":
        load_dir = Path(load_dir)
        df = pd.read_csv(load_dir / "df.csv")
        embs = np.load(load_dir / "embs.npy")
        df["embs"] = list(embs)
        return Sim(df, api_key=api_key, model=model)

    @staticmethod
    def check_load(
        load_dir: Union[str, Path],
        df: pd.DataFrame,
    ) -> bool:
        load_dir = Path(load_dir)
        if (
            not Path(load_dir / "df.csv").exists()
            or not Path(load_dir / "embs.npy").exists()
        ):
            return False

        df_load = pd.read_csv(load_dir / "df.csv")
        if platform.system() == "Windows":
            df_load = df_load.assign(
                doc=df_load.doc.apply(lambda x: x.replace("\r", ""))
            )
        return df.equals(df_load)  # type: ignore

    @lru_cache(maxsize=256)
    def top_k(
        self,
        query: str,
        k: int = 5,
        thresh: Optional[float] = None,
    ) -> Sequence[Dict]:
        """Returns the top k most similar items to the query.

        Parameters:
            query: str: The query to compare to.
            k: int: The number of items to return.
            thresh: Optional[float]: The minimum similarity threshold.

        Returns:
            Sequence[Dict]: The top k most similar items.
        """

        embedding = get_embedding(
            self.emb_call,
            query,
        )
        self.df = self.df.assign(
            sim=self.df.embs.apply(lambda x: 1 - cosine(x, embedding))
        )
        res = self.df.sort_values("sim", ascending=False).head(k)
        if thresh is not None:
            res = res[res.sim > thresh]
        return res[[c for c in res.columns if c != "embs"]].to_dict(orient="records")


class AzureSim(Sim):
    def __init__(
        self,
        df: pd.DataFrame,
        sim_key: Optional[str] = None,
        api_key: Optional[str] = None,
        api_version: str = "2024-02-01",
        azure_endpoint: Optional[str] = None,
        model: Optional[str] = None,
    ) -> None:
        if not api_key:
            api_key = os.getenv("AZURE_OPENAI_API_KEY")
        if not azure_endpoint:
            azure_endpoint = os.getenv("AZURE_OPENAI_ENDPOINT")
        if not model:
            model = os.getenv("AZURE_OPENAI_EMBEDDING_MODEL_DEPLOYMENT_NAME")

        if not api_key:
            raise ValueError("Azure OpenAI API key is required.")
        if not azure_endpoint:
            raise ValueError("Azure OpenAI endpoint is required.")
        if not model:
            raise ValueError(
                "Azure OpenAI embedding model deployment name is required."
            )

        self.df = df
        client = AzureOpenAI(
            api_key=api_key, api_version=api_version, azure_endpoint=azure_endpoint
        )
        self.emb_call = (
            lambda text: client.embeddings.create(input=text, model=model)
            .data[0]
            .embedding
        )

        self.model = model
        if "embs" not in df.columns and sim_key is None:
            raise ValueError("key is required if no column 'embs' is present.")

        if sim_key is not None:
            self.df = self.df.assign(
                embs=self.df[sim_key].apply(
                    lambda x: get_embedding(
                        self.emb_call,
                        x,
                    )
                )
            )


class OllamaSim(Sim):
    def __init__(
        self,
        df: pd.DataFrame,
        sim_key: Optional[str] = None,
        model_name: Optional[str] = None,
        base_url: Optional[str] = None,
    ) -> None:
        self.df = df
        if base_url is None:
            base_url = "http://localhost:11434/api/embeddings"
        if model_name is None:
            model_name = "mxbai-embed-large"

        def emb_call(text: List[str]) -> List[float]:
            resp = requests.post(
                base_url, json={"prompt": text[0], "model": model_name}
            )
            return resp.json()["embedding"]  # type: ignore

        self.emb_call = emb_call

        if "embs" not in df.columns and sim_key is None:
            raise ValueError("key is required if no column 'embs' is present.")

        if sim_key is not None:
            self.df = self.df.assign(
                embs=self.df[sim_key].apply(
                    lambda x: get_embedding(
                        self.emb_call,
                        x,
                    )
                )
            )


class StellaSim(Sim):
    def __init__(
        self,
        df: pd.DataFrame,
        sim_key: Optional[str] = None,
    ) -> None:
        self.df = df

        def emb_call(text: List[str]) -> List[float]:
            return stella_embeddings(text)[0]  # type: ignore

        self.emb_call = emb_call

        if "embs" not in df.columns and sim_key is None:
            raise ValueError("key is required if no column 'embs' is present.")

        if sim_key is not None:
            self.df = self.df.assign(
                embs=self.df[sim_key].apply(
                    lambda x: get_embedding(
                        self.emb_call,
                        x,
                    )
                )
            )

    @staticmethod
    def load(
        load_dir: Union[str, Path],
        api_key: Optional[str] = None,
        model: str = "stella1.5b",
    ) -> "StellaSim":
        load_dir = Path(load_dir)
        df = pd.read_csv(load_dir / "df.csv")
        embs = np.load(load_dir / "embs.npy")
        df["embs"] = list(embs)
        return StellaSim(df)


def merge_sim(sim1: Sim, sim2: Sim) -> Sim:
    return Sim(pd.concat([sim1.df, sim2.df], ignore_index=True))


def load_sim(sim_file: Union[str, Path]) -> Sim:
    sim_file = Path(sim_file)
    df = pd.read_csv(sim_file / "df.csv")
    embs = np.load(sim_file / "embs.npy")
    df["embs"] = list(embs)
    return Sim(df)



import copy
import logging
import os
import sys
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Union, cast

from tabulate import tabulate

import vision_agent.tools as T
from vision_agent.agent.agent import Agent
from vision_agent.agent.agent_utils import (
    _MAX_TABULATE_COL_WIDTH,
    DefaultImports,
    extract_code,
    extract_tag,
    format_feedback,
    print_code,
    remove_installs_from_code,
    strip_function_calls,
)
from vision_agent.agent.vision_agent_coder_prompts import (
    CODE,
    FIX_BUG,
    FULL_TASK,
    SIMPLE_TEST,
)
from vision_agent.agent.vision_agent_planner import (
    AnthropicVisionAgentPlanner,
    AzureVisionAgentPlanner,
    OllamaVisionAgentPlanner,
    OpenAIVisionAgentPlanner,
    PlanContext,
)
from vision_agent.lmm import (
    LMM,
    AnthropicLMM,
    AzureOpenAILMM,
    Message,
    OllamaLMM,
    OpenAILMM,
)
from vision_agent.tools.meta_tools import get_diff
from vision_agent.utils import CodeInterpreterFactory, Execution
from vision_agent.utils.execute import CodeInterpreter

logging.basicConfig(stream=sys.stdout)
WORKSPACE = Path(os.getenv("WORKSPACE", ""))
_LOGGER = logging.getLogger(__name__)


def write_code(
    coder: LMM,
    chat: List[Message],
    plan: str,
    tool_info: str,
    plan_thoughts: str,
    tool_output: str,
    feedback: str,
) -> str:
    chat = copy.deepcopy(chat)
    if chat[-1]["role"] != "user":
        raise ValueError("Last chat message must be from the user.")

    user_request = chat[-1]["content"]
    prompt = CODE.format(
        docstring=tool_info,
        question=FULL_TASK.format(user_request=user_request, subtasks=plan),
        tool_output=tool_output,
        plan_thoughts=plan_thoughts,
        feedback=feedback,
    )
    chat[-1]["content"] = prompt
    return extract_code(coder(chat, stream=False))  # type: ignore


def write_test(
    tester: LMM,
    chat: List[Message],
    tool_utils: str,
    code: str,
    feedback: str,
    media: Optional[Sequence[Union[str, Path]]] = None,
) -> str:
    chat = copy.deepcopy(chat)
    if chat[-1]["role"] != "user":
        raise ValueError("Last chat message must be from the user.")

    user_request = chat[-1]["content"]
    prompt = SIMPLE_TEST.format(
        docstring=tool_utils,
        question=user_request,
        code=code,
        feedback=feedback,
        media=media,
    )
    chat[-1]["content"] = prompt
    return extract_code(tester(chat, stream=False))  # type: ignore


def write_and_test_code(
    chat: List[Message],
    plan: str,
    tool_info: str,
    tool_output: str,
    plan_thoughts: str,
    tool_utils: str,
    working_memory: List[Dict[str, str]],
    coder: LMM,
    tester: LMM,
    debugger: LMM,
    code_interpreter: CodeInterpreter,
    log_progress: Callable[[Dict[str, Any]], None],
    verbosity: int = 0,
    max_retries: int = 3,
    media: Optional[Sequence[Union[str, Path]]] = None,
) -> Dict[str, Any]:
    log_progress(
        {
            "type": "log",
            "log_content": "Generating code",
            "status": "started",
        }
    )
    code = write_code(
        coder,
        chat,
        plan,
        tool_info,
        tool_output,
        plan_thoughts,
        format_feedback(working_memory),
    )
    code = strip_function_calls(code)
    test = write_test(
        tester, chat, tool_utils, code, format_feedback(working_memory), media
    )

    log_progress(
        {
            "type": "log",
            "log_content": "Running code",
            "status": "running",
            "code": DefaultImports.prepend_imports(code),
            "payload": {
                "test": test,
            },
        }
    )
    result = code_interpreter.exec_isolation(
        f"{DefaultImports.to_code_string()}\n{code}\n{test}"
    )
    log_progress(
        {
            "type": "log",
            "log_content": (
                "Code execution succeeded"
                if result.success
                else "Code execution failed"
            ),
            "status": "completed" if result.success else "failed",
            "code": DefaultImports.prepend_imports(code),
            "payload": {
                "test": test,
            },
        }
    )
    if verbosity == 2:
        print_code("Initial code and tests:", code, test)
        _LOGGER.info(
            f"Initial code execution result:\n{result.text(include_logs=True)}"
        )

    count = 0
    new_working_memory: List[Dict[str, str]] = []
    while not result.success and count < max_retries:
        if verbosity == 2:
            _LOGGER.info(f"Start debugging attempt {count + 1}")
        code, test, result = debug_code(
            working_memory,
            debugger,
            code_interpreter,
            tool_info,
            code,
            test,
            result,
            new_working_memory,
            log_progress,
            verbosity,
        )
        count += 1

    if verbosity >= 1:
        print_code("Final code and tests:", code, test)

    return {
        "code": code,
        "test": test,
        "success": result.success,
        "test_result": result,
        "working_memory": new_working_memory,
    }


def debug_code(
    working_memory: List[Dict[str, str]],
    debugger: LMM,
    code_interpreter: CodeInterpreter,
    tool_info: str,
    code: str,
    test: str,
    result: Execution,
    new_working_memory: List[Dict[str, str]],
    log_progress: Callable[[Dict[str, Any]], None],
    verbosity: int = 0,
) -> tuple[str, str, Execution]:
    log_progress(
        {
            "type": "log",
            "log_content": ("Debugging code"),
            "status": "started",
        }
    )

    fixed_code = None
    fixed_test = None
    thoughts = ""
    success = False
    count = 0
    while not success and count < 3:
        try:
            # LLMs write worse code when it's in JSON, so we have it write JSON
            # followed by code each wrapped in markdown blocks.
            fixed_code_and_test_str = debugger(
                FIX_BUG.format(
                    docstring=tool_info,
                    code=code,
                    tests=test,
                    # Because of the way we trace function calls the trace information
                    # ends up in the results. We don't want to show this info to the
                    # LLM so we don't include it in the tool_output_str.
                    result="\n".join(
                        result.text(include_results=False).splitlines()[-50:]
                    ),
                    feedback=format_feedback(working_memory + new_working_memory),
                ),
                stream=False,
            )
            fixed_code_and_test_str = cast(str, fixed_code_and_test_str)
            thoughts_tag = extract_tag(fixed_code_and_test_str, "thoughts")
            thoughts = thoughts_tag if thoughts_tag is not None else ""
            fixed_code = extract_tag(fixed_code_and_test_str, "code")
            fixed_test = extract_tag(fixed_code_and_test_str, "test")

            if fixed_code is None and fixed_test is None:
                success = False
            else:
                success = True

        except Exception as e:
            _LOGGER.exception(f"Error while extracting JSON: {e}")

        count += 1

    old_code = code
    old_test = test

    if fixed_code is not None and fixed_code.strip() != "":
        code = fixed_code
    if fixed_test is not None and fixed_test.strip() != "":
        test = fixed_test

    new_working_memory.append(
        {
            "code": f"{code}\n{test}",
            "feedback": thoughts,
            "edits": get_diff(f"{old_code}\n{old_test}", f"{code}\n{test}"),
        }
    )
    log_progress(
        {
            "type": "log",
            "log_content": ("Running code"),
            "status": "running",
            "code": DefaultImports.prepend_imports(code),
            "payload": {
                "test": test,
            },
        }
    )

    result = code_interpreter.exec_isolation(
        f"{DefaultImports.to_code_string()}\n{code}\n{test}"
    )
    log_progress(
        {
            "type": "log",
            "log_content": (
                "Code execution succeed" if result.success else "Code execution failed"
            ),
            "status": "completed" if result.success else "failed",
            "code": DefaultImports.prepend_imports(code),
            "payload": {
                "test": test,
                # "result": result.to_json(),
            },
        }
    )
    if verbosity == 2:
        print_code("Code and test after attempted fix:", code, test)
        _LOGGER.info(
            f"Reflection: {thoughts}\nCode execution result after attempted fix: {result.text(include_logs=True)}"
        )

    return code, test, result


class VisionAgentCoder(Agent):
    """Vision Agent Coder is an agentic framework that can output code based on a user
    request. It can plan tasks, retrieve relevant tools, write code, write tests and
    reflect on failed test cases to debug code. It is inspired by AgentCoder
    https://arxiv.org/abs/2312.13010 and Data Interpeter https://arxiv.org/abs/2402.18679

    Example
    -------
        >>> import vision_agent as va
        >>> agent = va.agent.VisionAgentCoder()
        >>> code = agent("What percentage of the area of the jar is filled with coffee beans?", media="jar.jpg")
    """

    def __init__(
        self,
        planner: Optional[Agent] = None,
        coder: Optional[LMM] = None,
        tester: Optional[LMM] = None,
        debugger: Optional[LMM] = None,
        verbosity: int = 0,
        report_progress_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        code_interpreter: Optional[Union[str, CodeInterpreter]] = None,
    ) -> None:
        """Initialize the Vision Agent Coder.

        Parameters:
            planner (Optional[Agent]): The planner model to use. Defaults to
                AnthropicVisionAgentPlanner.
            coder (Optional[LMM]): The coder model to use. Defaults to AnthropicLMM.
            tester (Optional[LMM]): The tester model to use. Defaults to AnthropicLMM.
            debugger (Optional[LMM]): The debugger model to use. Defaults to AnthropicLMM.
            verbosity (int): The verbosity level of the agent. Defaults to 0. 2 is the
                highest verbosity level which will output all intermediate debugging
                code.
            report_progress_callback (Optional[Callable[Dict[str, Any]]]): a callback
                to report the progress of the agent. This is useful for streaming logs
                in a web application where multiple VisionAgentCoder instances are
                running in parallel. This callback ensures that the progress are not
                mixed up.
            code_interpreter (Optional[Union[str, CodeInterpreter]]): For string values
                it can be one of: None, "local" or "e2b". If None, it will read from
                the environment variable "CODE_SANDBOX_RUNTIME". If a CodeInterpreter
                object is provided it will use that.
        """

        self.planner = (
            AnthropicVisionAgentPlanner(verbosity=verbosity)
            if planner is None
            else planner
        )
        self.coder = AnthropicLMM(temperature=0.0) if coder is None else coder
        self.tester = AnthropicLMM(temperature=0.0) if tester is None else tester
        self.debugger = AnthropicLMM(temperature=0.0) if debugger is None else debugger
        self.verbosity = verbosity
        if self.verbosity > 0:
            _LOGGER.setLevel(logging.INFO)

        self.report_progress_callback = report_progress_callback
        self.code_interpreter = code_interpreter

    def __call__(
        self,
        input: Union[str, List[Message]],
        media: Optional[Union[str, Path]] = None,
    ) -> str:
        """Generate code based on a user request.

        Parameters:
            input (Union[str, List[Message]]): A conversation in the format of
                [{"role": "user", "content": "describe your task here..."}] or a string
                of just the contents.
            media (Optional[Union[str, Path]]): The media file to be used in the task.

        Returns:
            str: The code output by the VisionAgentCoder.
        """

        if isinstance(input, str):
            input = [{"role": "user", "content": input}]
            if media is not None:
                input[0]["media"] = [media]
        code_and_context = self.generate_code(input)
        return code_and_context["code"]  # type: ignore

    def generate_code_from_plan(
        self,
        chat: List[Message],
        plan_context: PlanContext,
        code_interpreter: Optional[CodeInterpreter] = None,
    ) -> Dict[str, Any]:
        """Generates code and other intermediate outputs from a chat input and a plan.
        The plan includes:
            - plans: The plans generated by the planner.
            - best_plan: The best plan selected by the planner.
            - plan_thoughts: The thoughts of the planner, including any modifications
                to the plan.
            - tool_doc: The tool documentation for the best plan.
            - tool_output: The tool output from the tools used by the best plan.

        Parameters:
            chat (List[Message]): A conversation in the format of
                [{"role": "user", "content": "describe your task here..."}].
            plan_context (PlanContext): The context of the plan, including the plans,
                best_plan, plan_thoughts, tool_doc, and tool_output.

        Returns:
            Dict[str, Any]: A dictionary containing the code output by the
                VisionAgentCoder and other intermediate outputs. include:
                - status (str): Whether or not the agent completed or failed generating
                    the code.
                - code (str): The code output by the VisionAgentCoder.
                - test (str): The test output by the VisionAgentCoder.
                - test_result (Execution): The result of the test execution.
                - plans (Dict[str, Any]): The plans generated by the planner.
                - plan_thoughts (str): The thoughts of the planner.
                - working_memory (List[Dict[str, str]]): The working memory of the agent.
        """
        if not chat:
            raise ValueError("Chat cannot be empty.")

        # NOTE: each chat should have a dedicated code interpreter instance to avoid concurrency issues
        code_interpreter = (
            self.code_interpreter
            if self.code_interpreter is not None
            and not isinstance(self.code_interpreter, str)
            else CodeInterpreterFactory.new_instance(
                code_sandbox_runtime=self.code_interpreter,
            )
        )
        with code_interpreter:
            chat = copy.deepcopy(chat)
            media_list = []
            for chat_i in chat:
                if "media" in chat_i:
                    for media in chat_i["media"]:
                        chat_i["content"] += f" Media name {media}"  # type: ignore
                        media_list.append(str(media))

            int_chat = cast(
                List[Message],
                [
                    (
                        {
                            "role": c["role"],
                            "content": c["content"],
                            "media": c["media"],
                        }
                        if "media" in c
                        else {"role": c["role"], "content": c["content"]}
                    )
                    for c in chat
                ],
            )

            code = ""
            test = ""
            working_memory: List[Dict[str, str]] = []
            plan = plan_context.plans[plan_context.best_plan]
            tool_doc = plan_context.tool_doc
            tool_output_str = plan_context.tool_output
            plan_thoughts_str = str(plan_context.plan_thoughts)

            if self.verbosity >= 1:
                plan_fixed = [{"instructions": e} for e in plan["instructions"]]
                _LOGGER.info(
                    f"Picked best plan:\n{tabulate(tabular_data=plan_fixed, headers='keys', tablefmt='mixed_grid', maxcolwidths=_MAX_TABULATE_COL_WIDTH)}"
                )

            results = write_and_test_code(
                chat=[{"role": c["role"], "content": c["content"]} for c in int_chat],
                plan=f"\n{plan['thoughts']}\n-"
                + "\n-".join([e for e in plan["instructions"]]),
                tool_info=tool_doc,
                tool_output=tool_output_str,
                plan_thoughts=plan_thoughts_str,
                tool_utils=T.UTILITIES_DOCSTRING,
                working_memory=working_memory,
                coder=self.coder,
                tester=self.tester,
                debugger=self.debugger,
                code_interpreter=code_interpreter,
                log_progress=self.log_progress,
                verbosity=self.verbosity,
                media=media_list,
            )
            success = cast(bool, results["success"])
            code = remove_installs_from_code(cast(str, results["code"]))
            test = remove_installs_from_code(cast(str, results["test"]))
            working_memory.extend(results["working_memory"])
            execution_result = cast(Execution, results["test_result"])

            return {
                "status": "completed" if success else "failed",
                "code": DefaultImports.prepend_imports(code),
                "test": test,
                "test_result": execution_result,
                "plans": plan_context.plans,
                "plan_thoughts": plan_thoughts_str,
                "working_memory": working_memory,
            }

    def generate_code(
        self,
        chat: List[Message],
        test_multi_plan: bool = True,
        custom_tool_names: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """Generates code and other intermediate outputs from a chat input.

        Parameters:
            chat (List[Message]): A conversation in the format of
                [{"role": "user", "content": "describe your task here..."}].
            test_multi_plan (bool): Whether to test multiple plans or just the best plan.
            custom_tool_names (Optional[List[str]]): A list of custom tool names to use
                for the planner.

        Returns:
            Dict[str, Any]: A dictionary containing the code output by the
                VisionAgentCoder and other intermediate outputs. include:
                - status (str): Whether or not the agent completed or failed generating
                    the code.
                - code (str): The code output by the VisionAgentCoder.
                - test (str): The test output by the VisionAgentCoder.
                - test_result (Execution): The result of the test execution.
                - plans (Dict[str, Any]): The plans generated by the planner.
                - plan_thoughts (str): The thoughts of the planner.
                - working_memory (List[Dict[str, str]]): The working memory of the agent.
        """
        if not chat:
            raise ValueError("Chat cannot be empty.")

        # NOTE: each chat should have a dedicated code interpreter instance to avoid concurrency issues
        code_interpreter = (
            self.code_interpreter
            if self.code_interpreter is not None
            and not isinstance(self.code_interpreter, str)
            else CodeInterpreterFactory.new_instance(
                code_sandbox_runtime=self.code_interpreter,
            )
        )
        with code_interpreter:
            plan_context = self.planner.generate_plan(  # type: ignore
                chat,
                test_multi_plan=test_multi_plan,
                custom_tool_names=custom_tool_names,
                code_interpreter=code_interpreter,
            )

            code_and_context = self.generate_code_from_plan(
                chat,
                plan_context,
                code_interpreter=code_interpreter,
            )
        return code_and_context

    def chat(self, chat: List[Message]) -> List[Message]:
        chat = copy.deepcopy(chat)
        code = self.generate_code(chat)
        chat.append({"role": "agent", "content": code["code"]})
        return chat

    def log_progress(self, data: Dict[str, Any]) -> None:
        if self.report_progress_callback is not None:
            self.report_progress_callback(data)


class OpenAIVisionAgentCoder(VisionAgentCoder):
    """Initializes Vision Agent Coder using OpenAI models for planning, coding, testing."""

    def __init__(
        self,
        planner: Optional[Agent] = None,
        coder: Optional[LMM] = None,
        tester: Optional[LMM] = None,
        debugger: Optional[LMM] = None,
        verbosity: int = 0,
        report_progress_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        code_interpreter: Optional[Union[str, CodeInterpreter]] = None,
    ) -> None:
        self.planner = (
            OpenAIVisionAgentPlanner(verbosity=verbosity)
            if planner is None
            else planner
        )
        self.coder = OpenAILMM(temperature=0.0) if coder is None else coder
        self.tester = OpenAILMM(temperature=0.0) if tester is None else tester
        self.debugger = OpenAILMM(temperature=0.0) if debugger is None else debugger
        self.verbosity = verbosity
        if self.verbosity > 0:
            _LOGGER.setLevel(logging.INFO)

        self.report_progress_callback = report_progress_callback
        self.code_interpreter = code_interpreter


class AnthropicVisionAgentCoder(VisionAgentCoder):
    """Initializes Vision Agent Coder using Anthropic models for planning, coding, testing."""

    def __init__(
        self,
        planner: Optional[Agent] = None,
        coder: Optional[LMM] = None,
        tester: Optional[LMM] = None,
        debugger: Optional[LMM] = None,
        verbosity: int = 0,
        report_progress_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        code_interpreter: Optional[Union[str, CodeInterpreter]] = None,
    ) -> None:
        # NOTE: Claude doesn't have an official JSON mode
        self.planner = (
            AnthropicVisionAgentPlanner(verbosity=verbosity)
            if planner is None
            else planner
        )
        self.coder = AnthropicLMM(temperature=0.0) if coder is None else coder
        self.tester = AnthropicLMM(temperature=0.0) if tester is None else tester
        self.debugger = AnthropicLMM(temperature=0.0) if debugger is None else debugger
        self.verbosity = verbosity
        if self.verbosity > 0:
            _LOGGER.setLevel(logging.INFO)

        self.report_progress_callback = report_progress_callback
        self.code_interpreter = code_interpreter


class OllamaVisionAgentCoder(VisionAgentCoder):
    """VisionAgentCoder that uses Ollama models for planning, coding, testing.

    Pre-requisites:
    1. Run ollama pull llama3.2-vision for the LMM
    2. Run ollama pull mxbai-embed-large for the embedding similarity model

    Example
    -------
        >>> image vision_agent as va
        >>> agent = va.agent.OllamaVisionAgentCoder()
        >>> code = agent("What percentage of the area of the jar is filled with coffee beans?", media="jar.jpg")
    """

    def __init__(
        self,
        planner: Optional[Agent] = None,
        coder: Optional[LMM] = None,
        tester: Optional[LMM] = None,
        debugger: Optional[LMM] = None,
        verbosity: int = 0,
        report_progress_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        code_interpreter: Optional[Union[str, CodeInterpreter]] = None,
    ) -> None:
        super().__init__(
            planner=(
                OllamaVisionAgentPlanner(verbosity=verbosity)
                if planner is None
                else planner
            ),
            coder=(
                OllamaLMM(model_name="llama3.2-vision", temperature=0.0)
                if coder is None
                else coder
            ),
            tester=(
                OllamaLMM(model_name="llama3.2-vision", temperature=0.0)
                if tester is None
                else tester
            ),
            debugger=(
                OllamaLMM(model_name="llama3.2-vision", temperature=0.0)
                if debugger is None
                else debugger
            ),
            verbosity=verbosity,
            report_progress_callback=report_progress_callback,
            code_interpreter=code_interpreter,
        )


class AzureVisionAgentCoder(VisionAgentCoder):
    """VisionAgentCoder that uses Azure OpenAI APIs for planning, coding, testing.

    Pre-requisites:
    1. Set the environment variable AZURE_OPENAI_API_KEY to your Azure OpenAI API key.
    2. Set the environment variable AZURE_OPENAI_ENDPOINT to your Azure OpenAI endpoint.

    Example
    -------
        >>> import vision_agent as va
        >>> agent = va.agent.AzureVisionAgentCoder()
        >>> code = agent("What percentage of the area of the jar is filled with coffee beans?", media="jar.jpg")
    """

    def __init__(
        self,
        planner: Optional[Agent] = None,
        coder: Optional[LMM] = None,
        tester: Optional[LMM] = None,
        debugger: Optional[LMM] = None,
        verbosity: int = 0,
        report_progress_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        code_interpreter: Optional[Union[str, CodeInterpreter]] = None,
    ) -> None:
        """Initialize the Vision Agent Coder.

        Parameters:
            planner (Optional[Agent]): The planner model to use. Defaults to
                AzureVisionAgentPlanner.
            coder (Optional[LMM]): The coder model to use. Defaults to OpenAILMM.
            tester (Optional[LMM]): The tester model to use. Defaults to OpenAILMM.
            debugger (Optional[LMM]): The debugger model to
            verbosity (int): The verbosity level of the agent. Defaults to 0. 2 is the
                highest verbosity level which will output all intermediate debugging
                code.
            report_progress_callback: a callback to report the progress of the agent.
                This is useful for streaming logs in a web application where multiple
                VisionAgentCoder instances are running in parallel. This callback
                ensures that the progress are not mixed up.
        """
        super().__init__(
            planner=(
                AzureVisionAgentPlanner(verbosity=verbosity)
                if planner is None
                else planner
            ),
            coder=AzureOpenAILMM(temperature=0.0) if coder is None else coder,
            tester=AzureOpenAILMM(temperature=0.0) if tester is None else tester,
            debugger=(
                AzureOpenAILMM(temperature=0.0) if debugger is None else debugger
            ),
            verbosity=verbosity,
            report_progress_callback=report_progress_callback,
            code_interpreter=code_interpreter,
        )



from pathlib import Path
from typing import Dict, Sequence, Union

from vision_agent.utils.execute import Execution

TextOrImage = Union[str, Sequence[Union[str, Path]]]
Message = Dict[str, Union[TextOrImage, Execution]]



import base64
import logging
import tempfile
from functools import lru_cache
from typing import List, Optional, Tuple

import av  # type: ignore
import cv2
import numpy as np

_LOGGER = logging.getLogger(__name__)
# The maximum length of the clip to extract frames from, in seconds

_DEFAULT_VIDEO_FPS = 24
_DEFAULT_INPUT_FPS = 1.0


def play_video(video_base64: str) -> None:
    """Play a video file"""
    video_data = base64.b64decode(video_base64)
    with tempfile.NamedTemporaryFile(delete=False, suffix=".mp4") as temp_video:
        temp_video.write(video_data)
        temp_video_path = temp_video.name

        cap = cv2.VideoCapture(temp_video_path)
        if not cap.isOpened():
            _LOGGER.error("Error: Could not open video.")
            return

        # Display the first frame and wait for any key press to start the video
        ret, frame = cap.read()
        if ret:
            cv2.imshow("Video Player", frame)
            _LOGGER.info(f"Press any key to start playing the video: {temp_video_path}")
            cv2.waitKey(0)  # Wait for any key press

        while cap.isOpened():
            ret, frame = cap.read()
            if not ret:
                break
            cv2.imshow("Video Player", frame)
            # Press 'q' to exit the video
            if cv2.waitKey(200) & 0xFF == ord("q"):
                break
        cap.release()
        cv2.destroyAllWindows()


def _resize_frame(frame: np.ndarray) -> np.ndarray:
    height, width = frame.shape[:2]
    new_width = width - (width % 2)
    new_height = height - (height % 2)
    return cv2.resize(frame, (new_width, new_height))


def video_writer(
    frames: List[np.ndarray],
    fps: float = _DEFAULT_INPUT_FPS,
    filename: Optional[str] = None,
) -> str:
    if isinstance(fps, str):
        # fps could be a string when it's passed in from a web endpoint deployment
        fps = float(fps)
    if filename is None:
        filename = tempfile.NamedTemporaryFile(delete=False, suffix=".mp4").name
    container = av.open(filename, mode="w")
    stream = container.add_stream("h264", rate=fps)
    height, width = frames[0].shape[:2]
    stream.height = height - (height % 2)
    stream.width = width - (width % 2)
    stream.pix_fmt = "yuv420p"
    stream.options = {"crf": "10"}
    for frame in frames:
        # Remove the alpha channel (convert RGBA to RGB)
        frame_rgb = frame[:, :, :3]
        # Resize the frame to make dimensions divisible by 2
        frame_rgb = _resize_frame(frame_rgb)
        av_frame = av.VideoFrame.from_ndarray(frame_rgb, format="rgb24")
        for packet in stream.encode(av_frame):
            container.mux(packet)

    for packet in stream.encode():
        container.mux(packet)
    container.close()
    return filename


def frames_to_bytes(
    frames: List[np.ndarray], fps: float = _DEFAULT_INPUT_FPS, file_ext: str = ".mp4"
) -> bytes:
    r"""Convert a list of frames to a video file encoded into a byte string.

    Parameters:
        frames: the list of frames
        fps: the frames per second of the video
        file_ext: the file extension of the video file
    """
    if isinstance(fps, str):
        # fps could be a string when it's passed in from a web endpoint deployment
        fps = float(fps)
    with tempfile.NamedTemporaryFile(delete=True, suffix=file_ext) as temp_file:
        video_writer(frames, fps, temp_file.name)

        with open(temp_file.name, "rb") as f:
            buffer_bytes = f.read()
    return buffer_bytes


# WARNING: This cache is a little dangerous because if the underlying video
# contents change but the filename remains the same it will return the old file contents.
# For vision agent it's unlikely to change the file contents while keeping the
# same file name and the time savings are very large.
@lru_cache(maxsize=8)
def extract_frames_from_video(
    video_uri: str, fps: float = _DEFAULT_INPUT_FPS
) -> List[Tuple[np.ndarray, float]]:
    """Extract frames from a video along with the timestamp in seconds.

    Parameters:
        video_uri (str): the path to the video file or a video file url
        fps (float): the frame rate per second to extract the frames

    Returns:
        a list of tuples containing the extracted frame and the timestamp in seconds.
            E.g. [(frame1, 0.0), (frame2, 0.5), ...]. The timestamp is the time in seconds
            from the start of the video. E.g. 12.125 means 12.125 seconds from the start of
            the video. The frames are sorted by the timestamp in ascending order.
    """
    if isinstance(fps, str):
        # fps could be a string when it's passed in from a web endpoint deployment
        fps = float(fps)

    cap = cv2.VideoCapture(video_uri)
    orig_fps = cap.get(cv2.CAP_PROP_FPS)
    if not orig_fps or orig_fps <= 0:
        _LOGGER.warning(
            f"Input video, {video_uri}, has no fps, using the default value {_DEFAULT_VIDEO_FPS}"
        )
        orig_fps = _DEFAULT_VIDEO_FPS
    if not fps or fps <= 0:
        _LOGGER.warning(
            f"Input fps, {fps}, is illegal, using the default value: {_DEFAULT_INPUT_FPS}"
        )
        fps = _DEFAULT_INPUT_FPS
    orig_frame_time = 1 / orig_fps
    targ_frame_time = 1 / fps
    frames: List[Tuple[np.ndarray, float]] = []
    i = 0
    elapsed_time = 0.0
    while cap.isOpened():
        ret, frame = cap.read()
        if not ret:
            break

        elapsed_time += orig_frame_time
        # This is to prevent float point precision loss issue, which can cause
        # the elapsed time to be slightly less than the target frame time, which
        # causes the last frame to be skipped
        elapsed_time = round(elapsed_time, 8)
        if elapsed_time >= targ_frame_time:
            frames.append((cv2.cvtColor(frame, cv2.COLOR_BGR2RGB), i / orig_fps))
            elapsed_time -= targ_frame_time

        i += 1
    cap.release()
    _LOGGER.info(f"Extracted {len(frames)} frames from {video_uri}")
    return frames



import copy
import logging
from json import JSONDecodeError
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

from pydantic import BaseModel
from tabulate import tabulate

import vision_agent.tools as T
from vision_agent.agent import Agent
from vision_agent.agent.agent_utils import (
    _MAX_TABULATE_COL_WIDTH,
    DefaultImports,
    extract_code,
    extract_json,
    format_feedback,
    format_plans,
    print_code,
)
from vision_agent.agent.vision_agent_planner_prompts import (
    PICK_PLAN,
    PLAN,
    PREVIOUS_FAILED,
    TEST_PLANS,
    USER_REQ,
)
from vision_agent.lmm import (
    LMM,
    AnthropicLMM,
    AzureOpenAILMM,
    Message,
    OllamaLMM,
    OpenAILMM,
)
from vision_agent.utils.execute import (
    CodeInterpreter,
    CodeInterpreterFactory,
    Execution,
)
from vision_agent.utils.sim import AzureSim, OllamaSim, Sim

_LOGGER = logging.getLogger(__name__)


class PlanContext(BaseModel):
    plans: Dict[str, Dict[str, Union[str, List[str]]]]
    best_plan: str
    plan_thoughts: str
    tool_output: str
    tool_doc: str
    test_results: Optional[Execution]


def retrieve_tools(
    plans: Dict[str, Dict[str, Any]],
    tool_recommender: Sim,
    log_progress: Callable[[Dict[str, Any]], None],
    verbosity: int = 0,
) -> Dict[str, str]:
    log_progress(
        {
            "type": "log",
            "log_content": ("Retrieving tools for each plan"),
            "status": "started",
        }
    )
    tool_info = []
    tool_desc = []
    tool_lists: Dict[str, List[Dict[str, str]]] = {}
    for k, plan in plans.items():
        tool_lists[k] = []
        for task in plan["instructions"]:
            tools = tool_recommender.top_k(task, k=2, thresh=0.3)
            tool_info.extend([e["doc"] for e in tools])
            tool_desc.extend([e["desc"] for e in tools])
            tool_lists[k].extend(
                {"description": e["desc"], "documentation": e["doc"]} for e in tools
            )

    if verbosity == 2:
        tool_desc_str = "\n".join(set(tool_desc))
        _LOGGER.info(f"Tools Description:\n{tool_desc_str}")

    tool_lists_unique = {}
    for k in tool_lists:
        tool_lists_unique[k] = "\n\n".join(
            set(e["documentation"] for e in tool_lists[k])
        )
    all_tools = "\n\n".join(set(tool_info))
    tool_lists_unique["all"] = all_tools
    return tool_lists_unique


def _check_plan_format(plan: Dict[str, Any]) -> bool:
    if not isinstance(plan, dict):
        return False

    for k in plan:
        if "thoughts" not in plan[k] or "instructions" not in plan[k]:
            return False
        if not isinstance(plan[k]["instructions"], list):
            return False
    return True


def write_plans(
    chat: List[Message], tool_desc: str, working_memory: str, model: LMM
) -> Dict[str, Any]:
    chat = copy.deepcopy(chat)
    if chat[-1]["role"] != "user":
        raise ValueError("Last message in chat must be from user")

    user_request = chat[-1]["content"]
    context = USER_REQ.format(user_request=user_request)
    prompt = PLAN.format(
        context=context,
        tool_desc=tool_desc,
        feedback=working_memory,
    )
    chat[-1]["content"] = prompt
    plans = extract_json(model(chat, stream=False))  # type: ignore

    count = 0
    while not _check_plan_format(plans) and count < 3:
        _LOGGER.info("Invalid plan format. Retrying.")
        plans = extract_json(model(chat, stream=False))  # type: ignore
        count += 1
        if count == 3:
            raise ValueError("Failed to generate valid plans after 3 attempts.")
    return plans


def write_and_exec_plan_tests(
    plans: Dict[str, Any],
    tool_info: str,
    media: List[str],
    model: LMM,
    log_progress: Callable[[Dict[str, Any]], None],
    code_interpreter: CodeInterpreter,
    verbosity: int = 0,
    max_retries: int = 3,
) -> Tuple[str, Execution]:

    plan_str = format_plans(plans)
    prompt = TEST_PLANS.format(
        docstring=tool_info, plans=plan_str, previous_attempts="", media=media
    )

    code = extract_code(model(prompt, stream=False))  # type: ignore
    log_progress(
        {
            "type": "log",
            "log_content": "Executing code to test plans",
            "code": DefaultImports.prepend_imports(code),
            "status": "running",
        }
    )
    tool_output = code_interpreter.exec_isolation(DefaultImports.prepend_imports(code))
    # Because of the way we trace function calls the trace information ends up in the
    # results. We don't want to show this info to the LLM so we don't include it in the
    # tool_output_str.
    tool_output_str = tool_output.text(include_results=False).strip()

    if verbosity == 2:
        print_code("Initial code and tests:", code)
        _LOGGER.info(f"Initial code execution result:\n{tool_output_str}")

    log_progress(
        {
            "type": "log",
            "log_content": (
                "Code execution succeeded"
                if tool_output.success
                else "Code execution failed"
            ),
            "code": DefaultImports.prepend_imports(code),
            # "payload": tool_output.to_json(),
            "status": "completed" if tool_output.success else "failed",
        }
    )

    # retry if the tool output is empty or code fails
    count = 0
    tool_output_str = tool_output.text(include_results=False).strip()
    while (
        not tool_output.success
        or (len(tool_output.logs.stdout) == 0 and len(tool_output.logs.stderr) == 0)
    ) and count < max_retries:
        prompt = TEST_PLANS.format(
            docstring=tool_info,
            plans=plan_str,
            previous_attempts=PREVIOUS_FAILED.format(
                code=code, error="\n".join(tool_output_str.splitlines()[-50:])
            ),
            media=media,
        )
        log_progress(
            {
                "type": "log",
                "log_content": "Retrying code to test plans",
                "status": "running",
                "code": DefaultImports.prepend_imports(code),
            }
        )
        code = extract_code(model(prompt, stream=False))  # type: ignore
        tool_output = code_interpreter.exec_isolation(
            DefaultImports.prepend_imports(code)
        )
        log_progress(
            {
                "type": "log",
                "log_content": (
                    "Code execution succeeded"
                    if tool_output.success
                    else "Code execution failed"
                ),
                "code": DefaultImports.prepend_imports(code),
                # "payload": tool_output.to_json(),
                "status": "completed" if tool_output.success else "failed",
            }
        )
        tool_output_str = tool_output.text(include_results=False).strip()

        if verbosity == 2:
            print_code("Code and test after attempted fix:", code)
            _LOGGER.info(f"Code execution result after attempt {count + 1}")
            _LOGGER.info(f"{tool_output_str}")

        count += 1

    return code, tool_output


def write_plan_thoughts(
    chat: List[Message],
    plans: Dict[str, Any],
    tool_output_str: str,
    model: LMM,
    max_retries: int = 3,
) -> Dict[str, str]:
    user_req = chat[-1]["content"]
    context = USER_REQ.format(user_request=user_req)
    # because the tool picker model gets the image as well, we have to be careful with
    # how much text we send it, so we truncate the tool output to 20,000 characters
    prompt = PICK_PLAN.format(
        context=context,
        plans=format_plans(plans),
        tool_output=tool_output_str[:20_000],
    )
    chat[-1]["content"] = prompt
    count = 0

    plan_thoughts = None
    while plan_thoughts is None and count < max_retries:
        try:
            plan_thoughts = extract_json(model(chat, stream=False))  # type: ignore
        except JSONDecodeError as e:
            _LOGGER.exception(
                f"Error while extracting JSON during picking best plan {str(e)}"
            )
            pass
        count += 1

    if (
        plan_thoughts is None
        or "best_plan" not in plan_thoughts
        or ("best_plan" in plan_thoughts and plan_thoughts["best_plan"] not in plans)
    ):
        _LOGGER.info(f"Failed to pick best plan. Using the first plan. {plan_thoughts}")
        plan_thoughts = {"best_plan": list(plans.keys())[0]}

    if "thoughts" not in plan_thoughts:
        plan_thoughts["thoughts"] = ""
    return plan_thoughts


def pick_plan(
    chat: List[Message],
    plans: Dict[str, Any],
    tool_info: str,
    model: LMM,
    code_interpreter: CodeInterpreter,
    media: List[str],
    log_progress: Callable[[Dict[str, Any]], None],
    verbosity: int = 0,
    max_retries: int = 3,
) -> Tuple[Dict[str, str], str, Execution]:
    log_progress(
        {
            "type": "log",
            "log_content": "Generating code to pick the best plan",
            "status": "started",
        }
    )

    chat = copy.deepcopy(chat)
    if chat[-1]["role"] != "user":
        raise ValueError("Last chat message must be from the user.")

    code, tool_output = write_and_exec_plan_tests(
        plans,
        tool_info,
        media,
        model,
        log_progress,
        code_interpreter,
        verbosity,
        max_retries,
    )

    if verbosity >= 1:
        print_code("Final code:", code)

    plan_thoughts = write_plan_thoughts(
        chat,
        plans,
        tool_output.text(include_results=False).strip(),
        model,
        max_retries,
    )

    if verbosity >= 1:
        _LOGGER.info(f"Best plan:\n{plan_thoughts}")
    log_progress(
        {
            "type": "log",
            "log_content": "Picked best plan",
            "status": "completed",
            "payload": plans[plan_thoughts["best_plan"]],
        }
    )
    return plan_thoughts, code, tool_output


class VisionAgentPlanner(Agent):
    def __init__(
        self,
        planner: Optional[LMM] = None,
        tool_recommender: Optional[Sim] = None,
        verbosity: int = 0,
        report_progress_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        code_interpreter: Optional[Union[str, CodeInterpreter]] = None,
    ) -> None:
        self.planner = AnthropicLMM(temperature=0.0) if planner is None else planner
        self.verbosity = verbosity
        if self.verbosity > 0:
            _LOGGER.setLevel(logging.INFO)

        self.tool_recommender = (
            Sim(T.TOOLS_DF, sim_key="desc")
            if tool_recommender is None
            else tool_recommender
        )
        self.report_progress_callback = report_progress_callback
        self.code_interpreter = code_interpreter

    def __call__(
        self, input: Union[str, List[Message]], media: Optional[Union[str, Path]] = None
    ) -> str:
        if isinstance(input, str):
            input = [{"role": "user", "content": input}]
            if media is not None:
                input[0]["media"] = [media]
        planning_context = self.generate_plan(input)
        return str(planning_context.plans[planning_context.best_plan])

    def generate_plan(
        self,
        chat: List[Message],
        test_multi_plan: bool = True,
        custom_tool_names: Optional[List[str]] = None,
        code_interpreter: Optional[CodeInterpreter] = None,
    ) -> PlanContext:
        if not chat:
            raise ValueError("Chat cannot be empty")

        code_interpreter = (
            code_interpreter
            if code_interpreter is not None
            else (
                self.code_interpreter
                if not isinstance(self.code_interpreter, str)
                else CodeInterpreterFactory.new_instance(self.code_interpreter)
            )
        )
        code_interpreter = cast(CodeInterpreter, code_interpreter)
        with code_interpreter:
            chat = copy.deepcopy(chat)
            media_list = []
            for chat_i in chat:
                if "media" in chat_i:
                    for media in chat_i["media"]:
                        chat_i["content"] += f" Media name {media}"  # type: ignore
                        media_list.append(str(media))

            int_chat = cast(
                List[Message],
                [
                    (
                        {
                            "role": c["role"],
                            "content": c["content"],
                            "media": c["media"],
                        }
                        if "media" in c
                        else {"role": c["role"], "content": c["content"]}
                    )
                    for c in chat
                ],
            )

            working_memory: List[Dict[str, str]] = []

            plans = write_plans(
                chat,
                T.get_tool_descriptions_by_names(
                    custom_tool_names, T.FUNCTION_TOOLS, T.UTIL_TOOLS  # type: ignore
                ),
                format_feedback(working_memory),
                self.planner,
            )
            if self.verbosity >= 1:
                for plan in plans:
                    plan_fixed = [
                        {"instructions": e} for e in plans[plan]["instructions"]
                    ]
                    _LOGGER.info(
                        f"\n{tabulate(tabular_data=plan_fixed, headers='keys', tablefmt='mixed_grid', maxcolwidths=_MAX_TABULATE_COL_WIDTH)}"
                    )

            tool_docs = retrieve_tools(
                plans,
                self.tool_recommender,
                self.log_progress,
                self.verbosity,
            )
            if test_multi_plan:
                plan_thoughts, code, tool_output = pick_plan(
                    int_chat,
                    plans,
                    tool_docs["all"],
                    self.planner,
                    code_interpreter,
                    media_list,
                    self.log_progress,
                    self.verbosity,
                )
                best_plan = plan_thoughts["best_plan"]
                plan_thoughts_str = plan_thoughts["thoughts"]
                tool_output_str = (
                    "```python\n"
                    + code
                    + "\n```\n"
                    + tool_output.text(include_results=False).strip()
                )
            else:
                best_plan = list(plans.keys())[0]
                tool_output_str = ""
                plan_thoughts_str = ""
                tool_output = None

            if best_plan in plans and best_plan in tool_docs:
                tool_doc = tool_docs[best_plan]
            else:
                if self.verbosity >= 1:
                    _LOGGER.warning(
                        f"Best plan {best_plan} not found in plans or tool_infos. Using the first plan and tool info."
                    )
                k = list(plans.keys())[0]
                best_plan = k
                tool_doc = tool_docs[k]

        return PlanContext(
            plans=plans,
            best_plan=best_plan,
            plan_thoughts=plan_thoughts_str,
            tool_output=tool_output_str,
            test_results=tool_output,
            tool_doc=tool_doc,
        )

    def log_progress(self, log: Dict[str, Any]) -> None:
        if self.report_progress_callback is not None:
            self.report_progress_callback(log)


class AnthropicVisionAgentPlanner(VisionAgentPlanner):
    def __init__(
        self,
        planner: Optional[LMM] = None,
        tool_recommender: Optional[Sim] = None,
        verbosity: int = 0,
        report_progress_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        code_interpreter: Optional[Union[str, CodeInterpreter]] = None,
    ) -> None:
        super().__init__(
            planner=AnthropicLMM(temperature=0.0) if planner is None else planner,
            tool_recommender=tool_recommender,
            verbosity=verbosity,
            report_progress_callback=report_progress_callback,
            code_interpreter=code_interpreter,
        )


class OpenAIVisionAgentPlanner(VisionAgentPlanner):
    def __init__(
        self,
        planner: Optional[LMM] = None,
        tool_recommender: Optional[Sim] = None,
        verbosity: int = 0,
        report_progress_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        code_interpreter: Optional[Union[str, CodeInterpreter]] = None,
    ) -> None:
        super().__init__(
            planner=(OpenAILMM(temperature=0.0) if planner is None else planner),
            tool_recommender=tool_recommender,
            verbosity=verbosity,
            report_progress_callback=report_progress_callback,
            code_interpreter=code_interpreter,
        )


class OllamaVisionAgentPlanner(VisionAgentPlanner):
    def __init__(
        self,
        planner: Optional[LMM] = None,
        tool_recommender: Optional[Sim] = None,
        verbosity: int = 0,
        report_progress_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        code_interpreter: Optional[Union[str, CodeInterpreter]] = None,
    ) -> None:
        super().__init__(
            planner=(
                OllamaLMM(model_name="llama3.2-vision", temperature=0.0)
                if planner is None
                else planner
            ),
            tool_recommender=(
                OllamaSim(T.TOOLS_DF, sim_key="desc")
                if tool_recommender is None
                else tool_recommender
            ),
            verbosity=verbosity,
            report_progress_callback=report_progress_callback,
            code_interpreter=code_interpreter,
        )


class AzureVisionAgentPlanner(VisionAgentPlanner):
    def __init__(
        self,
        planner: Optional[LMM] = None,
        tool_recommender: Optional[Sim] = None,
        verbosity: int = 0,
        report_progress_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        code_interpreter: Optional[Union[str, CodeInterpreter]] = None,
    ) -> None:
        super().__init__(
            planner=(AzureOpenAILMM(temperature=0.0) if planner is None else planner),
            tool_recommender=(
                AzureSim(T.TOOLS_DF, sim_key="desc")
                if tool_recommender is None
                else tool_recommender
            ),
            verbosity=verbosity,
            report_progress_callback=report_progress_callback,
            code_interpreter=code_interpreter,
        )



VA_CODE = """
**Role**: You are a helpful agent that assists users with writing code.

**Taks**: As a conversational agent, you are required to understand the user's request and provide a helpful response. Use a Chain-of-Thought approach to break down the problem, create a plan, and then provide a response. Ensure that your response is clear, concise, and helpful. You can use an interactive Python (Jupyter Notebook) environment, executing code with <execute_python>. You are given access to an `artifacts` object which contains files shared between you and the user. `artifacts` will be saved only AFTER you execute python code with `save_imgae` or `save_video`. The user can see all `artifacts`.

<execute_python>
print("Hello World!")
</execute_python>

**Documentation**:
This is the documentation for the different actions you can take:

{documentation}

**Examples**:
Here is an example of how you can interact with a user and Actions to complete a task:
--- START EXAMPLES ---
{examples}
--- END EXAMPLES ---

**Conversation**:
Here is the current conversation so far:
--- START CONVERSATION ---
{conversation}
--- END CONVERSATION ---

**Instructions**:
1. **Understand and Clarify**: Make sure you understand the task, ask clarifying questions if the task is not clear.
2. **Code Generation**: Only use code provided in the Documentation in your <execute_python> tags. Only use `edit_vision_code` to modify code written by `generate_vision_code`. DO NOT run `edit_vision_code` or `edit_code_artifact` more than 2 times in a row and instead ask the user for help.
3. **Execute**: Do only what the user asked you to do and no more. If you need to ask the user a question or show your results to the user, set <let_user_respond> to `true`.
4. **Artifacts**: Files are only saved in `artifacts` after <execute_python>, do not try to access artifacts until you observe that they are loaded.
5. **Response**: Keep your responses short and concise. Provide the user only with the information they need to continue the conversation.
6. **Output**: You can only respond with <thinking>, <response>, <execute_python>, and <let_user_respond> tags.

<thinking>Your thoughts here...</thinking>
<response>Your response to the user here...</response>
<execute_python>Your code here...</execute_python>
<let_user_respond>true/false for whether or not to you want to let the user respond.</let_user_respond>
"""


EXAMPLES_CODE1 = """
USER: Can you write code to detect the dogs in this image? Media name dog.jpg

OBSERVATION:
[Artifacts loaded]
Artifact dog.jpg loaded to /path/to/images/dog.jpg
[End of artifacts]

AGENT: <thinking>I will use the generate_vision_code to detect the dogs in the image.</thinking>
<execute_python>
generate_vision_code(artifacts, 'dog_detector.py', 'Can you write code to detect dogs in this image?', media=['/path/to/images/dog.jpg'])
</execute_python>
<let_user_respond>false</let_user_respond>

OBSERVATION:
[Artifact dog_detector.py (5 lines total)]
0|from vision_agent.tools import load_image, owlv2_object_detection
1|def detect_dogs(image_path: str):
2|    image = load_image(image_path)
3|    dogs = owlv2_object_detection("dog", image)
4|    return dogs
[End of artifact]

AGENT: <thinking>I have generated the code to detect the dogs in the image, I must now run the code and print the results to get the output.</thinking>
<execute_python>
from dog_detector import detect_dogs
print(detect_dogs('/example/workspace/dog.jpg'))
</execute_python>
<let_user_respond>false</let_user_respond>

OBSERVATION:
----- stdout -----
[{'score': 0.99, 'label': 'dog', 'box': [0.1, 0.2, 0.3, 0.4]}, {'score': 0.23, 'label': 'dog', 'box': [0.2, 0.3, 0.4, 0.5]}]


AGENT: <thinking>Two dogs are detected, I will show this to the user and ask them if the result looks good.</thinking>
<response>The code detectd two dogs, do the results look good to you?</response>
<let_user_respond>true</let_user_respond>
"""

EXAMPLES_CODE1_EXTRA = """
USER: The the image only has one dog, can you fix this?

OBSERVATION:
[Artifacts loaded]
Artifact dog.jpg loaded to /path/to/images/dog.jpg
Artifact dog_detector.py loaded to /path/to/code/dog_detector.py
[End of artifacts]

AGENT: <thinking>I will use edit_vision_code to fix the code to detect only one dog. The lower score is 0.23 so I will set the threshold above that</thinking>
<execute_python>
edit_vision_code(artifacts, 'dog_detector.py', ['Can you write code to detect dogs in this image?', 'Can you increase the threshold to 0.24?'], media=['/path/to/images/dog.jpg'])
</execute_python>
<let_user_respond>false</let_user_respond>

OBSERVATION:
[Artifact dog_detector.py (5 lines total)]
0|from vision_agent.tools import load_image, owlv2_object_detection
1|def detect_dogs(image_path: str):
2|    image = load_image(image_path)
3|    dogs = owlv2_object_detection("dog", image, threshold=0.24)
4|    return dogs
[End of artifact]

AGENT: <thinking>I have edited the code to detect only one dog, I must now run the code and print the results to get the output.</thinking>
<execute_python>
from dog_detector import detect_dogs
print(detect_dogs('/path/to/images/dog.jpg'))
</execute_python>
<let_user_respond>false</let_user_respond>

OBSERVATION:
----- stdout -----
[{'score': 0.99, 'label': 'dog', 'box': [0.1, 0.2, 0.3, 0.4]}]

AGENT: <thinking>One dog is detected, I will show this to the user and ask them if the result looks good.</thinking>
<response>The code detected one dog, do these results look good to you?</response>
<let_user_respond>true</let_user_respond>
"""

EXAMPLES_CODE2 = """
USER: Can you describe this image?

OBSERVATION:
[Artifacts loaded]
Artifact image.jpg loaded to /path/to/images/image.jpg
[End of artifacts]

AGENT: <thinking>The user hasn't asked me to write any code and the task is very simple so I will view the image and answer myself to respond to the user quickly.</thinking>
<execute_python>
view_media_artifacts('image.jpg')
</execute_python>
<let_user_respond>false</let_user_respond>

OBSERVATION:
[Image image.jpg displayed]

AGENT: <thinking>The image shows a cat and a dog sitting on the couch, I will tell the user and ask them if they need any other assistance.</thinking>
<response>The image contains a dog and a cat sitting on a couch. Can I help you with any other tasks?</response>
<let_user_respond>true</let_user_respond>
"""


EXAMPLES_CODE3 = """
USER: Can you create a function to count workers with helmets, return the count and a visaulization of the bounding boxes?

OBSERVATION:
[Artifacts loaded]
[End of artifacts]

AGENT: <thinking>The user has asked to count workers with helmets but has not provided an image. I will ask the user for an image and then generate the code to count workers with helmets.</thinking>
<response>Can you provide an image of workers with helmets?</response>
<let_user_respond>true</let_user_respond>

USER: Yes you can use workers.png

OBSERVATION:
[Artifacts loaded]
Artifact workers.png loaded to /path/to/images/workers.png
[End of artifacts]

AGENT: <thinking>I will use the generate_vision_code to count the workers with helmets in the image.</thinking>
<execute_python>
generate_vision_code(artifacts, 'code.py', 'Can you create a function to count workers with helmets, return the count and save a visaulization of the bounding boxes?', media=['/paths/to/images/workers.png'])
</execute_python>
<let_user_respond>false</let_user_respond>

OBSERVATION:
[Artifact code.py (17 lines total)]
0|from vision_agent.tools import load_image, florence2_object_detection, closest_box_distance, overlay_bounding_boxes, save_image
1|def count_workers_with_helmets(image_path: str):
2|    image = load_image(image_path)
3|    detections = florence2_object_detection("worker, helmet", image)
4|    workers = [d for d in detections if d['label'] == 'worker']
5|    helmets = [d for d in detections if d['label'] == 'helmet']
6|    count = 0
6|    for worker in workers:
7|        person_has_helmet = False
8|        for helmet in helmets:
9|            if closest_box_distance(worker['box'], helmet['box']) < 0.01:
10|                person_has_helmet = True
11|                break
11|        if person_has_helmet:
13|            count += 1
14|    viz = overlay_bounding_boxes(image, detections)
16|    return count, viz
[End of artifact]

AGENT: <thinking>I have generated the code to count the workers with helmets in the image, I must now run the code, print the counts and save the visualization.</thinking>
<execute_python>
from code import count_workers_with_helmets
count, viz = count_workers_with_helmets('/path/to/images/workers.png')
save_image(viz, 'workers_viz.png')
print(count)
</execute_python>
<let_user_respond>false</let_user_respond>

OBSERVATION:
----- stdout -----
2

AGENT: <thinking>Two workers with helmets are detected, I will show this to the user and ask them if the result looks good.</thinking>
<response>The code to detect workers with helmets is saved in code.py and the visualization under 'workers_viz.png'.</response>
<let_user_respond>true</let_user_respond>
"""

EXAMPLES_CODE3_EXTRA1 = """
USER: The detections are slightly off. Can you fine tune florence2 using these labels? "[{'image_path': 'image1.png': 'labels': ['worker', 'helmet'], 'bboxes': [[235, 118, 294, 241], [232, 118, 294, 128]]}, {'image_path': 'image2.png': 'labels': ['worker', 'helmet'], 'bboxes': [[235, 118, 294, 241], [232, 118, 294, 128]]}]"

AGENT: <thinking>Because the user has supplied me with labels I can call object_detection_fine_tuning on their behalf to fine tune the model</thinking>
<execute_python>
object_detection_fine_tuning([{'image_path': 'image1.png': 'labels': ['worker', 'helmet'], 'bboxes': [[235, 118, 294, 241], [232, 118, 294, 128]]}, {'image_path': 'image2.png': 'labels': ['worker', 'helmet'], 'bboxes': [[235, 118, 294, 241], [232, 118, 294, 128]]}])
</execute_python>
<let_user_respond>false</let_user_respond>

OBSERVATION:
[Fine tuning id: 23b3b022-5ebf-4798-9373-20ef36429abf]

AGENT: <thinking>The model has finished fine tuning, I will now replace the original florence2_object_detection call with the fine tuning id.</thinking>
<execute_python>
use_object_detection_fine_tuning(artifacts, "code.py", "23b3b022-5ebf-4798-9373-20ef36429abf")
</execute_python>
<let_user_respond>false</let_user_respond>

OBSERVATION:
[Artifact code.py edits]
---
+++
@@ -1,7 +1,7 @@
 from vision_agent.tools import load_image, florence2_object_detection, closest_box_distance, overlay_bounding_boxes, save_image
 def count_workers_with_helmets(image_path: str, output_path: str):
     image = load_image(image_path)
-    detections = florence2_object_detection("worker, helmet", image)
+    detections = florence2_object_detection("worker, helmet", image, "23b3b022-5ebf-4798-9373-20ef36429abf")
     workers = [d for d in detections if d['label'] == 'worker']
     helmets = [d for d in detections if d['label'] == 'helmet']
     count = 0
[End of edits]

AGENT: <thinking>I have updated the code to use the fine tuned model, I must now run the code and print the output.</thinking>
<execute_python>
from code import count_workers_with_helmets
print(count_workers_with_helmets('/path/to/images/workers.png', 'workers_viz.png'))
</execute_python>
<let_user_respond>false</let_user_respond>

OBSERVATION:
----- stdout -----
3

AGENT: <thinking>Three workers with helmets are detected, I will show this to the user and ask them if the result looks good.</thinking>
<response>I have updated the code using the fine tuned florence2_object_detection model to count the workers wearing helmets in code.py and saved the visualization under 'workers_viz.png'.</response>
<let_user_respond>true</let_user_respond>
"""

EXAMPLES_CODE3_EXTRA2 = """
USER: Can you write a streamlit app that allows me to upload an image and count the workers with helmets?

AGENT: <thinking>I will use edit_code_artifact to generate a streamlit app that allows the user to upload an image and count the workers with helmets.</thinking>
<execute_python>
CODE = '''
import streamlit as st
from code imoprt count_workers_with_helmets
import os

def main():
    st.title("Worker with Helmet Counter")
    uploaded_file = st.file_uploader("Choose image...", type=["jpeg", "jpg", "png"])
    if uploaded_file is not None:
        with open("temp_image.jpg", "wb") as f:
            f.write(uploaded_file.read())

        count = count_workers_with_helmets("temp_image.jpg", "temp_image_viz.jpg")
        st.image("temp_image_viz.jpg")
        st.write(f"Number of workers with helmets: {count}")
        os.remove("temp_image.jpg")
        os.remove("temp_image_viz.jpg")

if __name__ == "__main__":
    main()
'''
edit_code_artifact(artifacts, 'streamlit_app.py', 0, 0, CODE)
</execute_python>

OBSERVATION:
[Artifact streamlit_app.py (19 lines total)]
0|import streamlit as st
1|from code imoprt count_workers_with_helmets
2|import os
3|
4|def main():
5|    st.title("Worker with Helmet Counter")
6|    uploaded_file = st.file_uploader("Choose image...", type=["jpeg", "jpg", "png"])
7|    if uploaded_file is not None:
8|        with open("temp_image.jpg", "wb") as f:
9|            f.write(uploaded_file.read())
10|
11|        count = count_workers_with_helmets("temp_image.jpg", "temp_image_viz.jpg")
12|        st.image("temp_image_viz.jpg")
13|        st.write(f"Number of workers with helmets: {count}")
14|        os.remove("temp_image.jpg")
15|        os.remove("temp_image_viz.jpg")
16|
17|if __name__ == "__main__":
18|    main()
[End of artifact]

AGENT: <thinking>I have generated the streamlit app that allows the user to upload an image and count the workers</thinking>
<response>The streamlit app is saved in streamlit_app.py, you can run the app by running `streamlit run streamlit_app.py`.</response>
<let_user_respond>true</let_user_respond>
"""



from pathlib import Path
from typing import List, Literal, Optional, Union

from pydantic import BaseModel

from vision_agent.utils.execute import Execution


class AgentMessage(BaseModel):
    """AgentMessage encompases messages sent to the entire Agentic system, which includes
    both LMMs and sub-agents.

    user: The user's message.
    assistant: The assistant's message.
    observation: An observation made after conducting an action, either by the user or
        assistant.
    interaction: An interaction between the user and the assistant. For example if the
        assistant wants to ask the user for help on a task, it could send an
        interaction message.
    interaction_response: The user's response to an interaction message.
    conversation: Messages coming from the conversation agent, this is a type of
        assistant messages.
    planner: Messages coming from the planner agent, this is a type of assistant
        messages.
    coder: Messages coming from the coder agent, this is a type of assistant messages.
    """

    role: Union[
        Literal["user"],
        Literal["assistant"],  # planner, coder and conversation are of type assistant
        Literal["observation"],
        Literal["interaction"],
        Literal["interaction_response"],
        Literal["conversation"],
        Literal["planner"],
        Literal["planner_update"],
        Literal["coder"],
    ]
    content: str
    media: Optional[List[Union[str, Path]]] = None


class PlanContext(BaseModel):
    """PlanContext is a data model that represents the context of a plan.

    plan: A description of the overall plan.
    instructions: A list of step-by-step instructions.
    code: Code snippets that were used during planning.
    """

    plan: str
    instructions: List[str]
    code: str


class CodeContext(BaseModel):
    """CodeContext is a data model that represents final code and test cases.

    code: The final code that was written.
    test: The test cases that were written.
    success: A boolean value indicating whether the code passed the test cases.
    test_result: The result of running the test cases.
    """

    code: str
    test: str
    success: bool
    test_result: Execution


class InteractionContext(BaseModel):
    """InteractionContext is a data model that represents the context of an interaction.

    chat: A list of messages exchanged between the user and the assistant.
    """

    chat: List[AgentMessage]



import copy
import json
import logging
import os
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

from vision_agent.agent import Agent
from vision_agent.agent.agent_utils import extract_json, extract_tag
from vision_agent.agent.vision_agent_prompts import (
    EXAMPLES_CODE1,
    EXAMPLES_CODE2,
    EXAMPLES_CODE3,
    EXAMPLES_CODE3_EXTRA2,
    VA_CODE,
)
from vision_agent.lmm import LMM, AnthropicLMM, Message, OpenAILMM
from vision_agent.tools.meta_tools import (
    META_TOOL_DOCSTRING,
    Artifacts,
    check_and_load_image,
    use_extra_vision_agent_args,
)
from vision_agent.utils import CodeInterpreterFactory
from vision_agent.utils.execute import CodeInterpreter, Execution

logging.basicConfig(level=logging.INFO)
_LOGGER = logging.getLogger(__name__)
WORKSPACE = Path(os.getenv("WORKSPACE", ""))
WORKSPACE.mkdir(parents=True, exist_ok=True)
if str(WORKSPACE) != "":
    os.environ["PYTHONPATH"] = f"{WORKSPACE}:{os.getenv('PYTHONPATH', '')}"


class BoilerplateCode:
    pre_code = [
        "from typing import *",
        "from vision_agent.utils.execute import CodeInterpreter",
        "from vision_agent.tools.meta_tools import Artifacts, open_code_artifact, create_code_artifact, edit_code_artifact, get_tool_descriptions, generate_vision_code, edit_vision_code, view_media_artifact, object_detection_fine_tuning, use_object_detection_fine_tuning, list_artifacts",
        "artifacts = Artifacts('{cwd}')",
    ]
    post_code: List[str] = []

    @staticmethod
    def add_boilerplate(code: str, **format: Any) -> str:
        """Run this method to prepend the default imports to the code.
        NOTE: be sure to run this method after the custom tools have been registered.
        """
        return (
            "\n".join([s.format(**format) for s in BoilerplateCode.pre_code])
            + "\n\n"
            + code
            + "\n\n"
            + "\n".join([s.format(**format) for s in BoilerplateCode.post_code])
        )


def format_agent_message(agent_message: str) -> str:
    agent_message_json = extract_json(agent_message)
    output = ""
    if "thinking" in agent_message_json and agent_message_json["thinking"]:
        output += "<thinking>" + agent_message_json["thinking"] + "</thinking>"
    if "response" in agent_message_json and agent_message_json["response"]:
        output += "<response>" + agent_message_json["response"] + "</response>"
    if "execute_python" in agent_message_json and agent_message_json["execute_python"]:
        output += (
            "\n<execute_python>\n"
            + agent_message_json["execute_python"]
            + "\n</execute_python>\n"
        )
    if (
        "let_user_respond" in agent_message_json
        and agent_message_json["let_user_respond"]
    ):
        output += (
            "<let_user_respond>"
            + str(agent_message_json["let_user_respond"])
            + "</let_user_respond>"
        )

    return output


def _clean_response(response: str) -> str:
    # Sometimes the LLM will hallucinate responses to an <execute_python> tag as if it
    # had already executed the code. This function removes the hallucinated response.
    if "<execute_python>" in response:
        end_execute_python = response.find("</execute_python>")
        response = response[: end_execute_python + len("</execute_python>")]
    return response


def run_conversation(orch: LMM, chat: List[Message]) -> Dict[str, Any]:
    chat = copy.deepcopy(chat)

    # only add 10 most recent messages in the chat to not go over token limit
    conversation = ""
    for chat_i in chat[-10:]:
        if chat_i["role"] == "user":
            conversation += f"USER: {chat_i['content']}\n\n"
        elif chat_i["role"] == "observation":
            conversation += f"OBSERVATION:\n{chat_i['content']}\n\n"
        elif chat_i["role"] == "assistant":
            conversation += f"AGENT: {format_agent_message(chat_i['content'])}\n\n"  # type: ignore
        else:
            raise ValueError(f"role {chat_i['role']} is not supported")

    prompt = VA_CODE.format(
        documentation=META_TOOL_DOCSTRING,
        examples=f"{EXAMPLES_CODE1}\n{EXAMPLES_CODE2}\n{EXAMPLES_CODE3}\n{EXAMPLES_CODE3_EXTRA2}",
        conversation=conversation,
    )
    message: Message = {"role": "user", "content": prompt}
    # only add recent media so we don't overload the model with old images
    if (
        chat[-1]["role"] == "observation"
        and "media" in chat[-1]
        and len(chat[-1]["media"]) > 0  # type: ignore
    ):
        media_obs = [media for media in chat[-1]["media"] if Path(media).exists()]  # type: ignore
        if len(media_obs) > 0:
            message["media"] = media_obs  # type: ignore
    conv_resp = cast(str, orch([message], stream=False))

    # clean the response first, if we are executing code, do not resond or end
    # conversation before the code has been executed.
    conv_resp = _clean_response(conv_resp)

    let_user_respond_str = extract_tag(conv_resp, "let_user_respond")
    let_user_respond = (
        "true" in let_user_respond_str.lower() if let_user_respond_str else False
    )

    return {
        "thinking": extract_tag(conv_resp, "thinking"),
        "response": extract_tag(conv_resp, "response"),
        "execute_python": extract_tag(conv_resp, "execute_python"),
        "let_user_respond": let_user_respond,
    }


def execute_code_action(
    artifacts: Artifacts,
    code: str,
    code_interpreter: CodeInterpreter,
) -> Tuple[Execution, str]:
    result = code_interpreter.exec_isolation(
        BoilerplateCode.add_boilerplate(code, cwd=str(artifacts.cwd))
    )

    obs = str(result.logs)
    if result.error:
        obs += f"\n{result.error}"
    return result, obs


def execute_user_code_action(
    artifacts: Artifacts,
    last_user_message: Message,
    code_interpreter: CodeInterpreter,
) -> Tuple[Optional[Execution], Optional[str]]:
    user_result = None
    user_obs = None

    if last_user_message["role"] != "user":
        return user_result, user_obs

    last_user_content = cast(str, last_user_message.get("content", ""))
    try:
        user_code_action = json.loads(last_user_content).get("execute_python", None)
    except json.JSONDecodeError:
        return user_result, user_obs

    if user_code_action is not None:
        user_code_action = use_extra_vision_agent_args(user_code_action, False)
        user_result, user_obs = execute_code_action(
            artifacts, user_code_action, code_interpreter
        )
        if user_result.error:
            user_obs += f"\n{user_result.error}"
    return user_result, user_obs


def add_step_descriptions(response: Dict[str, Any]) -> Dict[str, Any]:
    response = copy.deepcopy(response)

    if "execute_python" in response and response["execute_python"]:
        # only include descriptions for these, the rest will just have executing
        # code
        description_map = {
            "open_code_artifact": "Reading file.",
            "create_code_artifact": "Creating file.",
            "edit_code_artifact": "Editing file.",
            "generate_vision_code": "Generating vision code.",
            "edit_vision_code": "Editing vision code.",
        }
        description = ""
        for k, v in description_map.items():
            if k in response["execute_python"]:
                description += v + " "
        if description == "":
            description = "Executing code."

        response["response"] = description

    return response


def new_format_to_old_format(new_format: Dict[str, Any]) -> Dict[str, Any]:
    thoughts = new_format["thinking"] if new_format["thinking"] is not None else ""
    response = new_format["response"] if new_format["response"] is not None else ""
    if new_format["execute_python"] is not None:
        response += (
            f"\n<execute_python>\n{new_format['execute_python']}\n</execute_python>"
        )
    return {
        "thoughts": thoughts,
        "response": response,
        "let_user_respond": new_format["let_user_respond"],
    }


def old_format_to_new_format(old_format_str: str) -> str:
    try:
        old_format = json.loads(old_format_str)
    except json.JSONDecodeError:
        return old_format_str

    if "thoughts" in old_format:
        thinking = (
            old_format["thoughts"] if old_format["thoughts"].strip() != "" else None
        )
    else:
        thinking = None

    let_user_respond = (
        old_format["let_user_respond"] if "let_user_respond" in old_format else True
    )

    if "response" in old_format and "<execute_python>" in old_format["response"]:
        execute_python = extract_tag(old_format["response"], "execute_python")
        response = (
            old_format["response"]
            .replace(execute_python, "")
            .replace("<execute_python>", "")
            .replace("</execute_python>", "")
            .strip()
        )
    else:
        execute_python = None
        response = old_format["response"] if "response" in old_format else None

    return json.dumps(
        {
            "thinking": thinking,
            "response": response,
            "execute_python": execute_python,
            "let_user_respond": let_user_respond,
        }
    )


class VisionAgent(Agent):
    """Vision Agent is an agent that can chat with the user and call tools or other
    agents to generate code for it. Vision Agent uses python code to execute actions
    for the user. Vision Agent is inspired by by OpenDevin
    https://github.com/OpenDevin/OpenDevin and CodeAct https://arxiv.org/abs/2402.01030

    Example
    -------
        >>> from vision_agent.agent import VisionAgent
        >>> agent = VisionAgent()
        >>> resp = agent("Hello")
        >>> resp.append({"role": "user", "content": "Can you write a function that counts dogs?", "media": ["dog.jpg"]})
        >>> resp = agent(resp)
    """

    def __init__(
        self,
        agent: Optional[LMM] = None,
        cwd: Optional[Union[Path, str]] = None,
        verbosity: int = 0,
        callback_message: Optional[Callable[[Dict[str, Any]], None]] = None,
        code_sandbox_runtime: Optional[str] = None,
    ) -> None:
        """Initialize the VisionAgent.

        Parameters:
            agent (Optional[LMM]): The agent to use for conversation and orchestration
                of other agents.
            verbosity (int): The verbosity level of the agent.
            callback_message (Optional[Callable[[Dict[str, Any]], None]]): Callback
                function to send intermediate update messages.
            code_sandbox_runtime (Optional[str]): For string values it can be one of:
                None, "local" or "e2b". If None, it will read from the environment
                variable "CODE_SANDBOX_RUNTIME".
        """

        self.agent = AnthropicLMM(temperature=0.0) if agent is None else agent
        self.max_iterations = 12
        self.cwd = Path(cwd) if cwd is not None else Path.cwd()
        self.verbosity = verbosity
        self.callback_message = callback_message
        self.code_sandbox_runtime = code_sandbox_runtime
        if self.verbosity >= 1:
            _LOGGER.setLevel(logging.INFO)

    def __call__(
        self,
        input: Union[str, List[Message]],
        media: Optional[Union[str, Path]] = None,
        artifacts: Optional[Artifacts] = None,
    ) -> str:
        """Chat with VisionAgent and get the conversation response.

        Parameters:
            input (Union[str, List[Message]): A conversation in the format of
                [{"role": "user", "content": "describe your task here..."}, ...] or a
                string of just the contents.
            media (Optional[Union[str, Path]]): The media file to be used in the task.
            artifacts (Optional[Artifacts]): The artifacts to use in the task.

        Returns:
            str: The conversation response.
        """
        if isinstance(input, str):
            input = [{"role": "user", "content": input}]
            if media is not None:
                input[0]["media"] = [media]
        results, _ = self.chat_with_artifacts(input, artifacts)
        return results[-1]["content"]  # type: ignore

    def chat(
        self,
        chat: List[Message],
    ) -> List[Message]:
        """Chat with VisionAgent, it will use code to execute actions to accomplish
        its tasks.

        Parameters:
            chat (List[Message]): A conversation in the format of:
                [{"role": "user", "content": "describe your task here..."}]
                or if it contains media files, it should be in the format of:
                [{"role": "user", "content": "describe your task here...", "media": ["image1.jpg", "image2.jpg"]}]

        Returns:
            List[Message]: The conversation response.
        """
        return self.chat_with_artifacts(chat)[0]

    def chat_with_artifacts(
        self,
        chat: List[Message],
        artifacts: Optional[Artifacts] = None,
        test_multi_plan: bool = True,
        custom_tool_names: Optional[List[str]] = None,
    ) -> Tuple[List[Message], Artifacts]:
        """Chat with VisionAgent, it will use code to execute actions to accomplish
        its tasks.

        Parameters:
            chat (List[Message]): A conversation in the format of:
                [{"role": "user", "content": "describe your task here..."}]
                or if it contains media files, it should be in the format of:
                [{"role": "user", "content": "describe your task here...", "media": ["image1.jpg", "image2.jpg"]}]
            artifacts (Optional[Artifacts]): The artifacts to use in the task.
            test_multi_plan (bool): If True, it will test tools for multiple plans and
                pick the best one based off of the tool results. If False, it will go
                with the first plan.
            custom_tool_names (List[str]): A list of customized tools for agent to
                pick and use. If not provided, default to full tool set from
                vision_agent.tools.

        Returns:
            List[Message]: The conversation response.
        """

        if not chat:
            raise ValueError("chat cannot be empty")

        if not artifacts:
            artifacts = Artifacts(self.cwd)

        with CodeInterpreterFactory.new_instance(
            code_sandbox_runtime=self.code_sandbox_runtime,
            remote_path=self.cwd,
        ) as code_interpreter:
            orig_chat = copy.deepcopy(chat)
            int_chat = copy.deepcopy(chat)
            last_user_message = chat[-1]
            for chat_i in int_chat:
                if "media" in chat_i:
                    for media in chat_i["media"]:
                        media = cast(str, media)
                        media_remote_path = Path(artifacts.cwd) / Path(media).name
                        chat_i["content"] += f" Media name {media_remote_path}"  # type: ignore

            int_chat = cast(
                List[Message],
                [
                    (
                        {
                            "role": c["role"],
                            "content": old_format_to_new_format(c["content"]),  # type: ignore
                            "media": c["media"],
                        }
                        if "media" in c
                        else {"role": c["role"], "content": old_format_to_new_format(c["content"])}  # type: ignore
                    )
                    for c in int_chat
                ],
            )

            finished = False
            iterations = 0
            last_response = None

            # Upload artifacts to remote location and show where they are going
            # to be loaded to. The actual loading happens in BoilerplateCode as
            # part of the pre_code.
            artifacts_loaded = artifacts.show()
            int_chat.append({"role": "observation", "content": artifacts_loaded})
            orig_chat.append({"role": "observation", "content": artifacts_loaded})
            self.streaming_message({"role": "observation", "content": artifacts_loaded})

            user_result, user_obs = execute_user_code_action(
                artifacts,
                last_user_message,
                code_interpreter,
            )
            finished = user_result is not None and user_obs is not None
            if user_result is not None and user_obs is not None:
                # be sure to update the chat with user execution results
                chat_elt: Message = {"role": "observation", "content": user_obs}
                int_chat.append(chat_elt)
                chat_elt["execution"] = user_result
                orig_chat.append(chat_elt)
                self.streaming_message(
                    {
                        "role": "observation",
                        "content": user_obs,
                        "execution": user_result,
                        "finished": finished,
                    }
                )

            while not finished and iterations < self.max_iterations:
                response = run_conversation(self.agent, int_chat)
                if self.verbosity >= 1:
                    _LOGGER.info(response)

                code_action = response.get("execute_python", None)
                # sometimes it gets stuck in a loop, so we force it to exit
                if last_response == response:
                    response["let_user_respond"] = True
                    self.streaming_message(
                        {
                            "role": "assistant",
                            "content": "{}",
                            "error": {
                                "name": "Error when running conversation agent",
                                "value": "Agent is stuck in conversation loop, exited",
                                "traceback_raw": [],
                            },
                            "finished": True,
                        }
                    )
                else:
                    self.streaming_message(
                        {
                            "role": "assistant",
                            "content": new_format_to_old_format(
                                add_step_descriptions(response)
                            ),
                            "finished": response.get("let_user_respond", False)
                            and code_action is None,
                        }
                    )

                int_chat.append(
                    {
                        "role": "assistant",
                        "content": json.dumps(
                            new_format_to_old_format(add_step_descriptions(response))
                        ),
                    }
                )
                orig_chat.append(
                    {
                        "role": "assistant",
                        "content": json.dumps(
                            new_format_to_old_format(add_step_descriptions(response))
                        ),
                    }
                )
                finished = response.get("let_user_respond", False)

                if code_action is not None:
                    code_action = use_extra_vision_agent_args(
                        code_action, test_multi_plan, custom_tool_names
                    )

                if code_action is not None:
                    result, obs = execute_code_action(
                        artifacts,
                        code_action,
                        code_interpreter,
                    )
                    obs_chat_elt: Message = {"role": "observation", "content": obs}
                    media_obs = check_and_load_image(code_action)
                    if media_obs and result.success:
                        obs_chat_elt["media"] = [
                            artifacts.cwd / media_ob for media_ob in media_obs
                        ]

                    if self.verbosity >= 1:
                        _LOGGER.info(obs)

                    # don't add execution results to internal chat
                    int_chat.append(obs_chat_elt)
                    obs_chat_elt["execution"] = result
                    orig_chat.append(obs_chat_elt)
                    self.streaming_message(
                        {
                            "role": "observation",
                            "content": obs,
                            "execution": result,
                            "finished": finished,
                        }
                    )

                iterations += 1
                last_response = response

        return orig_chat, artifacts

    def streaming_message(self, message: Dict[str, Any]) -> None:
        if self.callback_message:
            self.callback_message(message)

    def log_progress(self, data: Dict[str, Any]) -> None:
        pass


class OpenAIVisionAgent(VisionAgent):
    def __init__(
        self,
        agent: Optional[LMM] = None,
        cwd: Optional[Union[Path, str]] = None,
        verbosity: int = 0,
        callback_message: Optional[Callable[[Dict[str, Any]], None]] = None,
    ) -> None:
        """Initialize the VisionAgent using OpenAI LMMs.

        Parameters:
            agent (Optional[LMM]): The agent to use for conversation and orchestration
                of other agents.
            verbosity (int): The verbosity level of the agent.
            callback_message (Optional[Callable[[Dict[str, Any]], None]]): Callback
                function to send intermediate update messages.
            code_interpreter (Optional[Union[str, CodeInterpreter]]): For string values
                it can be one of: None, "local" or "e2b". If None, it will read from
                the environment variable "CODE_SANDBOX_RUNTIME". If a CodeInterpreter
                object is provided it will use that.
        """

        agent = OpenAILMM(temperature=0.0, json_mode=True) if agent is None else agent
        super().__init__(
            agent,
            cwd,
            verbosity,
            callback_message,
        )


class AnthropicVisionAgent(VisionAgent):
    def __init__(
        self,
        agent: Optional[LMM] = None,
        cwd: Optional[Union[Path, str]] = None,
        verbosity: int = 0,
        callback_message: Optional[Callable[[Dict[str, Any]], None]] = None,
    ) -> None:
        """Initialize the VisionAgent using Anthropic LMMs.

        Parameters:
            agent (Optional[LMM]): The agent to use for conversation and orchestration
                of other agents.
            verbosity (int): The verbosity level of the agent.
            callback_message (Optional[Callable[[Dict[str, Any]], None]]): Callback
                function to send intermediate update messages.
            code_interpreter (Optional[Union[str, CodeInterpreter]]): For string values
                it can be one of: None, "local" or "e2b". If None, it will read from
                the environment variable "CODE_SANDBOX_RUNTIME". If a CodeInterpreter
                object is provided it will use that.
        """

        agent = AnthropicLMM(temperature=0.0) if agent is None else agent
        super().__init__(
            agent,
            cwd,
            verbosity,
            callback_message,
        )



from typing import Type

from pydantic import BaseModel, Field

from vision_agent.lmm import LMM, AnthropicLMM, OpenAILMM


class Config(BaseModel):
    # for vision_agent_v2
    agent: Type[LMM] = Field(default=AnthropicLMM)
    agent_kwargs: dict = Field(
        default_factory=lambda: {
            "model_name": "claude-3-5-sonnet-20241022",
            "temperature": 0.0,
            "image_size": 768,
        }
    )

    # for vision_agent_planner_v2
    planner: Type[LMM] = Field(default=AnthropicLMM)
    planner_kwargs: dict = Field(
        default_factory=lambda: {
            "model_name": "claude-3-5-sonnet-20241022",
            "temperature": 0.0,
            "image_size": 768,
        }
    )

    # for vision_agent_planner_v2
    summarizer: Type[LMM] = Field(default=OpenAILMM)
    summarizer_kwargs: dict = Field(
        default_factory=lambda: {
            "model_name": "o1",
            "temperature": 1.0,  # o1 has fixed temperature
            "image_size": 768,
        }
    )

    # for vision_agent_planner_v2
    critic: Type[LMM] = Field(default=AnthropicLMM)
    critic_kwargs: dict = Field(
        default_factory=lambda: {
            "model_name": "claude-3-5-sonnet-20241022",
            "temperature": 0.0,
            "image_size": 768,
        }
    )

    # for vision_agent_coder_v2
    coder: Type[LMM] = Field(default=AnthropicLMM)
    coder_kwargs: dict = Field(
        default_factory=lambda: {
            "model_name": "claude-3-5-sonnet-20241022",
            "temperature": 0.0,
            "image_size": 768,
        }
    )

    # for vision_agent_coder_v2
    tester: Type[LMM] = Field(default=AnthropicLMM)
    tester_kwargs: dict = Field(
        default_factory=lambda: {
            "model_name": "claude-3-5-sonnet-20241022",
            "temperature": 0.0,
            "image_size": 768,
        }
    )

    # for vision_agent_coder_v2
    debugger: Type[LMM] = Field(default=AnthropicLMM)
    debugger_kwargs: dict = Field(
        default_factory=lambda: {
            "model_name": "claude-3-5-sonnet-20241022",
            "temperature": 0.0,
            "image_size": 768,
        }
    )

    # for get_tool_for_task
    tool_tester: Type[LMM] = Field(default=AnthropicLMM)
    tool_tester_kwargs: dict = Field(
        default_factory=lambda: {
            "model_name": "claude-3-5-sonnet-20241022",
            "temperature": 0.0,
            "image_size": 768,
        }
    )

    # for get_tool_for_task
    tool_chooser: Type[LMM] = Field(default=AnthropicLMM)
    tool_chooser_kwargs: dict = Field(
        default_factory=lambda: {
            "model_name": "claude-3-5-sonnet-20241022",
            "temperature": 1.0,
            "image_size": 768,
        }
    )

    # for suggestions module
    suggester: Type[LMM] = Field(default=AnthropicLMM)
    suggester_kwargs: dict = Field(
        default_factory=lambda: {
            "model_name": "claude-3-5-sonnet-20241022",
            "temperature": 1.0,
            "image_size": 768,
        }
    )

    # for vqa module
    vqa: Type[LMM] = Field(default=AnthropicLMM)
    vqa_kwargs: dict = Field(
        default_factory=lambda: {
            "model_name": "claude-3-5-sonnet-20241022",
            "temperature": 0.0,
            "image_size": 768,
        }
    )

    def create_agent(self) -> LMM:
        return self.agent(**self.agent_kwargs)

    def create_planner(self) -> LMM:
        return self.planner(**self.planner_kwargs)

    def create_summarizer(self) -> LMM:
        return self.summarizer(**self.summarizer_kwargs)

    def create_critic(self) -> LMM:
        return self.critic(**self.critic_kwargs)

    def create_coder(self) -> LMM:
        return self.coder(**self.coder_kwargs)

    def create_tester(self) -> LMM:
        return self.tester(**self.tester_kwargs)

    def create_debugger(self) -> LMM:
        return self.debugger(**self.debugger_kwargs)

    def create_tool_tester(self) -> LMM:
        return self.tool_tester(**self.tool_tester_kwargs)

    def create_tool_chooser(self) -> LMM:
        return self.tool_chooser(**self.tool_chooser_kwargs)

    def create_suggester(self) -> LMM:
        return self.suggester(**self.suggester_kwargs)

    def create_vqa(self) -> LMM:
        return self.vqa(**self.vqa_kwargs)



PLAN = """
**Role**: You are an expert planning agent that can understand the user request and search for a plan to accomplish it.

**Task**: As a planning agent you are required to understand the user's request and search for a plan to accomplish it. Use Chain-of-Thought approach to break down the problem, create a plan, and then provide a response. Esnure your response is clear, concise, and helpful. You can use an interactive Pyton (Jupyter Notebok) environment, executing code with <execute_python>, each execution is a new cell so old code and outputs are saved.

**Documentation**: this is the documentation for the functions you can use to accomplish the task:
{tool_desc}

**Example Planning**: Here are some examples of how you can search for a plan, in the examples the user output is denoted by USER, your output is denoted by AGENT and the observations after your code execution are denoted by OBSERVATION:
{examples}

**Current Planning**:
--- START PLANNING ---
{planning}
--- END PLANNING ---

**Instructions**:
1. Read over the user request and context provided and output <thinking> tags to indicate your thought process. You can <count> number of turns to complete the user's request.
2. You can execute python code in the ipython notebook using <execute_python> tags. Only output one <execute_python> tag at a time.
3. Only output <finalize_plan> when you are done planning and want to end the planning process. DO NOT output <finalize_plan> with <execute_python> tags, only after OBSERVATION's.
4. Only load/save files from {media_list} unless you specifically saved the file previously.
5. Ensure you always call `suggestion` initially and `get_tool_for_task` to get the right tool for the subtask.
6. Calling `plt.imshow` or `save_image` will display the image to you so you can check your results. If you see an image after <execute_python> it's generated from your code.
7. DO NOT hard code the answer into your code, it should be dynamic and work for any similar request.
8. DO NOT over index on claude35_vqa, if tool output is close to claude35_vqa's output you do not need to improve the tool.
9. You can only respond in the following format with a single <thinking>, <execute_python> or <finalize_plan> tag:

<thinking>Your thought process...</thinking>
<execute_python>Your code here</execute_python>
<finalize_plan>Your final recommended plan</finalize_plan>
"""

EXAMPLE_PLAN1 = """
--- EXAMPLE1 ---
USER: Count the number of pedestrians in the image.
<count>10</count>

AGENT: <thinking>I need to gather more information, I will ask for a description of the image and a solution to the problem to work towards as well as a suggestion on how best to solve it.</thinking>
<execute_python>
image = load_image('drone.jpg')
claude35_vqa('Can you describe this image? How many pedestrians do you count in the image?', [image])
suggestion('How can I count the number of pedestrians in the image?', [image])
</execute_python>

OBSERVATION: [claude35_vqa output]
From this aerial view of a busy urban street, it's difficult to clearly see or count individual pedestrians. The image shows a bird's eye view of a city intersection with multiple lanes of traffic, parked cars, sidewalks, and some green spaces. While there may be people in cars or on the sidewalks, they are not distinctly visible from this height and perspective. The focus of the image is more on the urban infrastructure, traffic patterns, and overall city layout rather than on individuals.
[end of claude35_vqa output]

[suggestions]
[suggestion 0]
The image is very large and the items you need to detect are small.

Step 1: You should start by splitting the image into overlapping sections and runing the detection algorithm on each section:

def subdivide_image(image):
    height, width, _ = image.shape
    mid_height = height // 2
    mid_width = width // 2
    overlap_height = int(mid_height * 0.1)
    overlap_width = int(mid_width * 0.1)
    top_left = image[:mid_height + overlap_height, :mid_width + overlap_width, :]
    top_right = image[:mid_height + overlap_height, mid_width - overlap_width:, :]
    bottom_left = image[mid_height - overlap_height:, :mid_width + overlap_width, :]
    bottom_right = image[mid_height - overlap_height:, mid_width - overlap_width:, :]
    return [top_left, top_right, bottom_left, bottom_right]

get_tool_for_task('<your prompt here>', subdivide_image(image))

Step 2: Once you have the detections from each subdivided image, you will need to merge them back together to remove overlapping predictions, be sure to tranlate the offset back to the original image:

def bounding_box_match(b1: List[float], b2: List[float], iou_threshold: float = 0.1) -> bool:
    # Calculate intersection coordinates
    x1 = max(b1[0], b2[0])
    y1 = max(b1[1], b2[1])
    x2 = min(b1[2], b2[2])
    y2 = min(b1[3], b2[3])

    # Calculate intersection area
    if x2 < x1 or y2 < y1:
        return False  # No overlap

    intersection = (x2 - x1) * (y2 - y1)

    # Calculate union area
    area1 = (b1[2] - b1[0]) * (b1[3] - b1[1])
    area2 = (b2[2] - b2[0]) * (b2[3] - b2[1])
    union = area1 + area2 - intersection

    # Calculate IoU
    iou = intersection / union if union > 0 else 0

    return iou >= iou_threshold

def merge_bounding_box_list(detections):
    merged_detections = []
    for detection in detections:
        matching_box = None
        for i, other in enumerate(merged_detections):
            if bounding_box_match(detection["bbox"], other["bbox"]):
                matching_box = i
                break

        if matching_box is not None:
            # Keep the box with higher confidence score
            if detection["score"] > merged_detections[matching_box]["score"]:
                merged_detections[matching_box] = detection
        else:
            merged_detections.append(detection)

def sub_image_to_original(elt, sub_image_position, original_size):
    offset_x, offset_y = sub_image_position
    return {
        "label": elt["label"],
        "score": elt["score"],
        "bbox": [
            (elt["bbox"][0] + offset_x) / original_size[1],
            (elt["bbox"][1] + offset_y) / original_size[0],
            (elt["bbox"][2] + offset_x) / original_size[1],
            (elt["bbox"][3] + offset_y) / original_size[0],
        ],
    }

def normalized_to_unnormalized(elt, image_size):
    return {
        "label": elt["label"],
        "score": elt["score"],
        "bbox": [
            elt["bbox"][0] * image_size[1],
            elt["bbox"][1] * image_size[0],
            elt["bbox"][2] * image_size[1],
            elt["bbox"][3] * image_size[0],
        ],
    }

height, width, _ = image.shape
mid_width = width // 2
mid_height = height // 2

detection_from_subdivided_images = []
for i, sub_image in enumerate(subdivided_images):
    detections = <your detection function here>("pedestrian", sub_image)
    unnorm_detections = [
        normalized_to_unnormalized(
            detection, (sub_image.shape[0], sub_image.shape[1])
        )
        for detection in detections
    ]
    offset_x = i % 2 * (mid_width - int(mid_width * 0.1))
    offset_y = i // 2 * (mid_height - int(mid_height * 0.1))
    offset_detections = [
        sub_image_to_original(
            unnorm_detection, (offset_x, offset_y), (height, width)
        )
        for unnorm_detection in unnorm_detections
    ]
    detection_from_subdivided_images.extend(offset_detections)

detections = merge_bounding_box_list(detection_from_subdivided_images)
[end of suggestion 0]
[end of suggestion]
<count>9</count>

AGENT: <thinking>Following the suggested plan, I will subdivide the image into smaller parts and find the right tool on the smaller parts. I will also re-check with claude35_vqa on the subdivided images.</thinking>
<execute_python>
def subdivide_image(image):
    height, width, _ = image.shape
    mid_height = height // 2
    mid_width = width // 2
    overlap_height = int(mid_height * 0.1)
    overlap_width = int(mid_width * 0.1)
    top_left = image[:mid_height + overlap_height, :mid_width + overlap_width, :]
    top_right = image[:mid_height + overlap_height, mid_width - overlap_width:, :]
    bottom_left = image[mid_height - overlap_height:, :mid_width + overlap_width, :]
    bottom_right = image[mid_height - overlap_height:, mid_width - overlap_width:, :]
    return [top_left, top_right, bottom_left, bottom_right]

images = subdivide_image(image)
claude35_vqa('How many pedestrians do you see in these images?', images)
get_tool_for_task('Count the number of pedestrians in the images', images)
</execute_python>

OBSERVATION:
[claude35_vqa_output]
In these aerial images, I can see approximately 5-6 pedestrians walking in various locations, including near the retail signage are and along the sidewalks.
[end of claude35_vqa_output]

[get_tool_for_task output]
After examining the image, I can see it's an aerial view of a busy urban intersection with multiple lanes of traffic. There are numerous cars visible, each likely containing at least one person (the driver). While it's difficult to count exact individuals, I can estimate based on the number of vehicles. The countgd_object_detection tool is the best choice for this task as it can detect and count multiple instances of an object given a text prompt. I will use this tool with the prompt 'pedistrian' to count the number of pedestrians in the image.

Tool Documentation:
countgd_object_detection(prompt: str, image: numpy.ndarray, box_threshold: float = 0.23) -> List[Dict[str, Any]]:
    'countgd_object_detection' is a tool that can detect multiple instances of an
    object given a text prompt. It is particularly useful when trying to detect and
    count a large number of objects. You can optionally separate object names in the
    prompt with commas. It returns a list of bounding boxes with normalized
    coordinates, label names and associated confidence scores.

    Parameters:
        prompt (str): The object that needs to be counted.
        image (np.ndarray): The image that contains multiple instances of the object.
        box_threshold (float, optional): The threshold for detection. Defaults
            to 0.23.

    Returns:
        List[Dict[str, Any]]: A list of dictionaries containing the score, label, and
            bounding box of the detected objects with normalized coordinates between 0
            and 1 (xmin, ymin, xmax, ymax). xmin and ymin are the coordinates of the
            top-left and xmax and ymax are the coordinates of the bottom-right of the
            bounding box.

    Example
    -------
        >>> countgd_object_detection("flower", image)
        [
            {'score': 0.49, 'label': 'flower', 'bbox': [0.1, 0.11, 0.35, 0.4]},
        ]
[end of get_tool_for_task output]
<count>8</count>

AGENT: <thinking>I need to now merge the boxes from all region and use the countgd_object_detection tool with the prompt 'pedestrian' as suggested by get_tool_for_task.</thinking>
<execute_python>
def bounding_box_match(b1: List[float], b2: List[float], iou_threshold: float = 0.1) -> bool:
    # Calculate intersection coordinates
    x1 = max(b1[0], b2[0])
    y1 = max(b1[1], b2[1])
    x2 = min(b1[2], b2[2])
    y2 = min(b1[3], b2[3])

    # Calculate intersection area
    if x2 < x1 or y2 < y1:
        return False  # No overlap

    intersection = (x2 - x1) * (y2 - y1)

    # Calculate union area
    area1 = (b1[2] - b1[0]) * (b1[3] - b1[1])
    area2 = (b2[2] - b2[0]) * (b2[3] - b2[1])
    union = area1 + area2 - intersection

    # Calculate IoU
    iou = intersection / union if union > 0 else 0

    return iou >= iou_threshold

def merge_bounding_box_list(detections):
    merged_detections = []
    for detection in detections:
        matching_box = None
        for i, other in enumerate(merged_detections):
            if bounding_box_match(detection["bbox"], other["bbox"]):
                matching_box = i
                break

        if matching_box is not None:
            # Keep the box with higher confidence score
            if detection["score"] > merged_detections[matching_box]["score"]:
                merged_detections[matching_box] = detection
        else:
            merged_detections.append(detection)

detections = []
for region in subdivide_image(image):
    detections.extend(countgd_object_detection("pedestrian", region))
detections = merge_bounding_box_list(detections)
print(len(detections))
</execute_python>

OBSERVATION: 24
<count>7</count>

AGENT: <thinking>countgd_object_detection detected 24 pedestrians in the image by subdividing the image into smaller regions, this differs from claude35_vqa's answer of 5-6 pedestrians. I will use countgd_object_detection's answer since it is specifically designed to count objects.</thinking>
<finalize_plan>This plan successfully detected 24 pedestrians in the image by subdividing the image into smaller regions and using the countgd_object_detection tool on each region. The bounding boxes were then merged to count the unique pedestrians.</finalize_plan>
--- END EXAMPLE1 ---
"""

EXAMPLE_PLAN2 = """
--- EXAMPLE2 ---
USER: Can you identify which boxes have tape on them in this video?
<count>10</count>

AGENT: <thinking>I need to gather more information, I will ask for a description of the video and a solution to the problem to work towards as well as a suggestion on how best to solve it.</thinking>
<execute_python>
frames = extract_frames_and_timestamps('video.mp4', 1)
frames  = [f["frame"] for f in frames]
claude35_vqa('Can you describe this video? How many boxes are there and do they have tape on them?', frames[:5])
suggestion('How can I identify boxes with tape on them in the video?', frames[:5])
</execute_python>

OBSERVATION: [claude35_vqa output]
In this video, there appear to be 2 cardboard boxes moving along a roller conveyor system. The boxes are brown/kraft colored and appear to have tape sealing them. You can see clear packing tape on the top and sides of the boxes.
[end of claude35_vqa output]

[suggestions]
[suggestion 0]
You are trying to track objects in a video and identify features on those objects. You need to first get a tool that can track objects in a video, and then for each object find another tool to identify the features on the object. You can use the following code to help with this task:

track_predictions = <object tracking tool>("object", video_frames)


# Step 1: go through each frame and each prediction and extract the predicted bounding boxes as crops
obj_to_info = {{}}
for frame, frame_predictions in zip(video_frames, track_predictions):
    for obj in frame_predictions:
        if obj["label"] not in obj_to_info:
            obj_to_info[obj["label"]] = []
        height, width = frame.shape[:2]
        crop = frame[
            int(obj["bbox"][1] * height) : int(obj["bbox"][3] * height),
            int(obj["bbox"][0] * width) : int(obj["bbox"][2] * width),
            :,
        ]
        # For each crop use an object detection tool, VQA tool or classification tool to identify if the object contains the features you want
        detections = <object detection tool to find features>("<your feature here>", crop)
        obj_to_info[obj["label"]].extend(detections)

print(f"{len(obj_to_info)} objects tracked")

objects_with_info = set()
for infos in obj_to_info:
    for info in info:
        if info["label"] == "<your feature here>":
            objects_with_info.add(info)
            break

print(f"{len(objects_with_info)} objects with features found")
[end of suggestion 0]
[end of suggestions]
<count>9</count>

AGENT: <thinking>Following the suggested plan, I first need to identify a tool for tracking.</thinking>
<execute_python>
get_tool_for_task('Identify and track the boxes in the video', frames[:5])
</execute_python>

OBSERVATION:
[get_tool_for_task output]
For tracking boxes moving on a conveyor belt, we need a tool that can consistently track the same box across frames without losing it or double counting. Looking at the outputs: florence2_sam2_video_tracking successfully tracks the single box across all 5 frames, maintaining consistent tracking IDs and showing the box's movement along the conveyor and using the prompt 'box'.

Tool Documentation:
def florence2_sam2_video_tracking(prompt: str, frames: List[np.ndarray], chunk_length: Optional[int] = 10) -> List[List[Dict[str, Any]]]:
    'florence2_sam2_video_tracking' is a tool that can track and segment multiple
    objects in a video given a text prompt such as category names or referring
    expressions. The categories in the text prompt are separated by commas. It returns
    a list of bounding boxes, label names, masks and associated probability scores and
    is useful for tracking and counting without duplicating counts.

    Parameters:
        prompt (str): The prompt to ground to the video.
        frames (List[np.ndarray]): The list of frames to ground the prompt to.
        chunk_length (Optional[int]): The number of frames to re-run florence2 to find
            new objects.
        fine_tune_id (Optional[str]): If you have a fine-tuned model, you can pass the
            fine-tuned model ID here to use it.

    Returns:
        List[List[Dict[str, Any]]]: A list of list of dictionaries containing the
            label, segmentation mask and bounding boxes. The outer list represents each
            frame and the inner list is the entities per frame. The detected objects
            have normalized coordinates between 0 and 1 (xmin, ymin, xmax, ymax). xmin
            and ymin are the coordinates of the top-left and xmax and ymax are the
            coordinates of the bottom-right of the bounding box. The mask is binary 2D
            numpy array where 1 indicates the object and 0 indicates the background.
            The label names are prefixed with their ID represent the total count.

    Example
    -------
        >>> florence2_sam2_video_tracking("car, dinosaur", frames)
        [
            [
                {
                    'label': '0: dinosaur',
                    'bbox': [0.1, 0.11, 0.35, 0.4],
                    'mask': array([[0, 0, 0, ..., 0, 0, 0],
                        ...,
                        [0, 0, 0, ..., 0, 0, 0]], dtype=uint8),
                },
            ],
            ...
        ]
[end of get_tool_for_task output]
<count>8</count>

AGENT: <thinking>Now that I have the tool and prompt for tracking I need to find a tool to identify if there is tape on each of the boxes.</thinking>
<execute_python>
track_predictions = florence2_sam2_video_tracking("box", frames)

# only use first 5 frames to reduce processing time
crops = []
for frame, frame_predictions in zip(frames[:5], track_predictions[:5]):
    for obj in frame_predictions:
        height, width = frame.shape[:2]
        crop = frame[
            int(obj["bbox"][1] * height) : int(obj["bbox"][3] * height),
            int(obj["bbox"][0] * width) : int(obj["bbox"][2] * width),
            :,
        ]
        crops.append(crop)

# only use first 5 crops to reduce processing time
get_tool_for_task('Identify if there is tape on the boxes', crops[:3])

OBSERVATION:
[get_tool_for_task output]
owlv2_object_detection performed best with the prompt 'tape', as it specifically detected multiple instances of tape with localized bounding boxes, which matches what's visible in the images.

'owlv2_object_detection' is a tool that can detect and count multiple objects given a
text prompt such as category names or referring expressions on images. The categories
in text prompt are separated by commas. It returns a list of bounding boxes with
normalized coordinates, label names and associated probability scores.

Parameters:
    prompt (str): The prompt to ground to the image.
    image (np.ndarray): The image to ground the prompt to.
    box_threshold (float, optional): The threshold for the box detection. Defaults
        to 0.10.
    fine_tune_id (Optional[str]): If you have a fine-tuned model, you can pass the
        fine-tuned model ID here to use it.

Returns:
    List[Dict[str, Any]]: A list of dictionaries containing the score, label, and
        bounding box of the detected objects with normalized coordinates between 0
        and 1 (xmin, ymin, xmax, ymax). xmin and ymin are the coordinates of the
        top-left and xmax and ymax are the coordinates of the bottom-right of the
        bounding box.

Example
-------
    >>> owlv2_object_detection("car, dinosaur", image)
    [
        {'score': 0.99, 'label': 'dinosaur', 'bbox': [0.1, 0.11, 0.35, 0.4]},
        {'score': 0.98, 'label': 'car', 'bbox': [0.2, 0.21, 0.45, 0.5},
    ]
[end of get_tool_for_task output]
<count>7</count>

AGENT: <thinking>Now that I have the tools for tracking and identifying 'tape' on the boxes, I will use them to accomplish the user's request.</thinking>
<execute_python>
obj_to_info = {{}}
for frame, frame_predictions in zip(frames, track_predictions):
    for obj in frame_predictions:
        if obj["label"] not in obj_to_info:
            obj_to_info[obj["label"]] = []
        height, width = frame.shape[:2]
        crop = frame[
            int(obj["bbox"][1] * height) : int(obj["bbox"][3] * height),
            int(obj["bbox"][0] * width) : int(obj["bbox"][2] * width),
            :,
        ]
        detections = owlv2_object_detection("tape", crop)
        obj_to_info[obj["label"]].extend(detections)


print(f"{len(obj_to_info)} boxes tracked")
objects_with_tape = set()
for infos in obj_to_info:
    for info in info:
        if info["label"] == "tape":
            objects_with_tape.add(info)
            break

print(f"{len(objects_with_tape)} boxes with tape found")
</execute_python>

OBSERVATION:
3 boxes were tracked
2 boxes with tape found
<count>6</count>

AGENT: <thinking>After tracking the boxes and identifying tape on them, I found 3 boxes in the video and 2 of them had tape on them. I will finalize this plan.</thinking>
<finalize_plan>This plan successfully tracked the boxes in the video and identified tape on them. It found 3 boxes in the video and 2 of them had tape on them.</finalize_plan>
--- END EXAMPLE2 ---
"""

PICK_PLAN = """
**Role**: You are an expert evaluator that can understand user requests and evaluate different plans to determine the best one.

**Task**: You are given a planning chains by a planning agent complete with thoughts, python executions and observations. You are also given 3 different responses to add to the chain. Your task is to evaluate the different plans and determine the best one for the user request.

**Current Planning**:
--- START PLANNING ---
{planning}
--- END PLANNING ---

**Responses**:
--- RESPONSE1 ---
{response1}
--- END RESPONSE1 ---
--- RESPONSE2 ---
{response2}
--- END RESPONSE2 ---
--- RESPONSE3 ---
{response3}
--- END RESPONSE3 ---

**Instructions**:
1. Read over the planning chain and responses.
2. The agent can output <thinking>, <execute_python> or <finalize_plan> tags. Any code under OBSERVATION: has not been executed by the agent, only code under <execute_python> tags has been executed.
3. Rate each instruction on a scale of 1 to 10 based on how well it solves the user request and give your thoughts on why you rated it that way.
4. Return your results in the following JSON format inside <json> tags:

<json>
{{
    "response1": # int: rating for response1,
    "response1_thoughts": # str: your thoughts on response1,
    "response2": # int: rating for response2,
    "response2_thoughts": # str: your thoughts on response2,
    "response3": # int: rating for response3,
    "response3_thoughts": # str: your thoughts on response3
}}
</json>
"""

CATEGORIZE_TOOL_REQUEST = """
You are given a task: "{task}" from the user. You must extract the type of category this task belongs to, it can be one or more of the following:
- "object detection and counting" - detecting objects or counting objects from a text prompt in an image.
- "instance segmentation" - segmenting objects in an image given a text prompt.
- "classification" - classifying objects in an image given a text prompt.
- "segmentation" - segmenting objects in an image or video given a text prompt.
- "OCR" - extracting text from an image.
- "VQA" - answering questions about an image or video, can also be used for text extraction.
- "DocQA" - answering questions about a document or extracting information from a document.
- "video object tracking" - tracking objects in a video.
- "depth and pose estimation" - estimating the depth or pose of objects in an image.
- "temporal localization" - localizing the time period an event occurs in a video.
- "inpainting" - filling in masked parts of an image.

Return the category or categories (comma separated) inside tags <category># your categories here</category>. If you are unsure about a task, it is better to include more categories than less.
"""

TEST_TOOLS = """
**Role**: You are an expert software programming that specializes in tool testing.

**Task**: You are responsible for testing different tools on a set of images to determine the best tool for the user request.

**Tools**: This is the documentation for the functions you have access to. You may call any of these functions to help you complete the task. They are available through importing `from vision_agent.tools import *`.
{tool_docs}

**Previous Attempts**: You previously tried to execute this code and got the following error or no output (if empty that means you have not tried to previously execute code, if it is 'EMPTY' that means you have tried but got no output):
{previous_attempts}

**User Request**:
{user_request}

**Media**:
{media}

**Examples**:
{examples}

**Instructions**:
1. List all the tools under **Tools** and the user request. Write a program to load the media and call the most relevant tools in parallel and print it's output along with other relevant information.
2. Create a dictionary where the keys are the tool name and the values are the tool outputs. Remove numpy arrays from the printed dictionary.
3. Your test case MUST run only on the given images which are {media}
4. For video tracking, use chunk_length=1 and at least 3 frames to ensure the best results when evaluating the tool.
5. Use mutually exclusive categories for prompts such as 'person, car' and not 'person, athlete' to avoid over counting.
6. Print this final dictionary.
7. Output your code in the following format wrapped in <code> tags:
<code>
# Your code here
</code>
"""

TEST_TOOLS_EXAMPLE1 = """
--- EXAMPLE1 ---
**User Request**:
Count the number of pedestrians across all the images.

**Media**:
["image1.jpg", "image2.jpg", "image3.jpg", "image4.jpg"]

<code>
from concurrent.futures import ThreadPoolExecutor, as_completed
from vision_agent.tools import load_image, owlv2_object_detection, florence2_object_detection, countgd_object_detection

# process functions in a try catch so that if it fails it doesn't cause `as_completed` to hang
def process_owlv2(image_paths):
    try:
        results = []
        for image_path in image_paths:
            image = load_image(image_path)
            results.extend(owlv2_object_detection("person", image))
    except Exception as e:
        results = f"Encountered error when executing process_owlv2: {str(e)}"
    return results

def process_florence2(image_paths):
    try:
        results = []
        for image_path in image_paths:
            image = load_image(image_path)
            results.extend(florence2_object_detection("person", image))
    except Exception as e:
        results = f"Encountered error when executing process_florence2: {str(e)}"
    return results

def process_countgd(image_paths):
    try:
        results = []
        for image_path in image_paths:
            image = load_image(image_path)
            results.extend(countgd_object_detection("person", image))
    except Exception as e:
        results = f"Encountered error when executing process_countgd: {str(e)}"
    return results

image_paths = ["image1.jpg", "image2.jpg", "image3.jpg", "image4.jpg"]

with ThreadPoolExecutor() as executor:
    futures = {{
        executor.submit(process_owlv2, image_paths): "owlv2_object_detection",
        executor.submit(process_florence2, image_paths): "florence2_phrase_grounding",
        executor.submit(process_countgd, image_paths): "countgd_object_detection",
    }}

    final_results = {{}}
    for future in as_completed(futures):
        tool_name = futures[future]
        final_results[tool_name] = future.result()

print(final_results)
</code>
--- END EXAMPLE1 ---
"""

TEST_TOOLS_EXAMPLE2 = """
--- EXAMPLE2 ---
**User Request**:
Count the number of people in the video.

**Media**:
["video.mp4"]

<code>
import numpy as np
from concurrent.futures import ThreadPoolExecutor, as_completed
from vision_agent.tools import extract_frames_and_timestamps, owlv2_sam2_video_tracking, florence2_sam2_video_tracking

# sample at 1 FPS and use the first 10 frames to reduce processing time
frames = extract_frames_and_timestamps("video.mp4", 1)
frames = [f["frame"] for f in frames][:10]

# strip arrays from the output to make it easier to read
def remove_arrays(o):
    if isinstance(o, list):
        return [remove_arrays(e) for e in o]
    elif isinstance(o, dict):
        return {{k: remove_arrays(v) for k, v in o.items()}}
    elif isinstance(o, np.ndarray):
        return "array: " + str(o.shape)
    else:
        return o

def process_owlv2_sam2_video_tracking(frames):
    try:
        # run with chunk_length=1 to ensure best results
        results = owlv2_sam2_video_tracking("person", frames, chunk_length=1)
    except Exception as e:
        results = f"Encountered error when executing process_owlv2_sam2_video_tracking: {str(e)}"
    return results

def process_florence2_sam2_video_tracking(frames):
    try:
        # run with chunk_length=1 to ensure best results
        results = florence2_sam2_video_tracking("person", frames, chunk_length=1)
    except Exception as e:
        results = f"Encountered error when executing process_florence2_sam2: {str(e)}"
    return results


with ThreadPoolExecutor() as executor:
    futures = {{
        executor.submit(process_owlv2_sam2_video_tracking, frames): "owlv2_sam2_video_tracking",
        executor.submit(process_florence2_sam2_video_tracking, frames): "florence2_sam2_video_tracking",
    }}
    final_results = {{}}
    for future in as_completed(futures):
        tool_name = futures[future]
        final_results[tool_name] = remove_arrays(future.result())

print(final_results)
</code>
--- END EXAMPLE2 ---
"""

PICK_TOOL = """
**Role**: You are an expert evaluator that can understand user requests and evaluate the output of different tools.

**Task**: You are given the output of different tools for a user request along with the image. You must evaluate the output and determine the best tool for the user request.

**User Request**:
{user_request}

**Tools**: This is the documentation of all the functions that were tested.
{tool_docs}

**Testing Code and Tool Output**:
{context}

**Previous Attempt**: This is the code and output of the previous attempt, if it is empty then there was no previous attempt.
{previous_attempts}

**Instructions**:
1. Re-read the user request, plans, tool outputs and examine the image.
2. Given the user request, try to solve the problem yourself.
3. Pick which tool output best matches your solution and the user request, DO NOT focus on other factors.
4. DO NOT modify confidence thresholds unless the tool output is completely wrong.
5. Remember for videos that in order to count objects a video some sort of tracking is needed, or else you will overcount the objects.
7. Return the following JSON format inside <json> tags using the exact tool name for best_tool:
<json>
{{
    "predicted_answer": str # the answer you would expect from the best plan
    "thoughts": str # your thought process for choosing the best tool over other tools and any modifications you madeas well as the prompt you used with the tool.
    "best_tool": str # the name of the best tool
}}
</json>
"""

FINALIZE_PLAN = """
**Task**: You are given a chain of thoughts, python executions and observations from a planning agent as it tries to construct a plan to solve a user request. Your task is to summarize the plan it found so that another programming agent to write a program to accomplish the user request.

**Documentation**: You can use these tools to help you visualize or save the output:
{tool_desc}

**Planning**: Here is chain of thoughts, executions and observations from the planning agent:
{planning}

**Instructions**:
1. Summarize the plan that the planning agent found.
2. Write a single function that solves the problem based on what the planner found and only returns the final solution.
3. Only use tools obtained from calling `get_tool_for_task`.
4. Do not include {excluded_tools} tools in your instructions.
5. Ensure the function is well documented and easy to understand.
6. Ensure you visualize the output with `overlay_bounding_boxes` or `overlay_segmentation_masks` and save it to a file with `save_image` or `save_video`.
7. Use the default FPS for extracting frames from videos unless otherwise specified by the user.
8. Include the expected answer in your 'plan' so that the programming agent can properly test if it has the correct answer.
9. Respond in the following format with JSON surrounded by <json> tags and code surrounded by <code> tags:

<json>
{{
    "plan": str # the plan you have summarized
    "instructions": [
        str # the instructions for each step in the plan with either specific code or a specific tool name
    ]
}}
</json>

<code>
# Code snippets here
</code>
"""

FIX_BUG = """
**Role**: You are a software programmer responsible for fixing bugs in code.

**Task**: Your responsibility is to fix the bug in the code and provide the correct output.

**Instructions**: This is the previous chat history:
--- CHAT HISTORY ---
{chat_history}
--- END CHAT HISTORY ---

This is the code that was previously run:
<python>
{code}
</python>

And this is the error message that was received:
<error>
{error}
</error>

Please fix the code by correcting the error. Do not wrap failures in try-except statements and instead solve the root cause. Output your code in the following format wrapped in <code> tags:

<code>
# Your fixed code here
</code>
"""

CRITIQUE_PLAN = """
**Role**: You are an expert evaluator that can understand user requests and evaluate different plans to determine the best one.

**Task**: You are given a planning chain by a planning agent complete with thoughts, python executions and observations. The agent is trying to use tools to accomplish a vision task requested by a user. Your job is to evaluate the plan so far given the user request and image.


**Current Planning**:
--- START PLANNING ---
{planning}
--- END PLANNING ---

**Instructions**:
1. Read over the planning chain and examine the image and user request. Try to solve the user problem yourself and compare it against the solution the agent has come up with so far.
2. Rate the plan on a scale of 1 to 10 based on how well it solves the user request. A score of 8+ means the plan is good and the agent should continue on the same path, 5-7 means the plan is good but needs some minor modifications, less than 5 means the plan is bad and the agent should consider a different approach.
3. Provide your score and thoughts in the following format:

<score>Your score here</score>
<thoughts>Your thoughts here</thoughts>
"""
